from __future__ import annotations
import typing
import attr
from .ode import ODE


@attr.s(frozen=True, slots=True)
class CodeComponent:
    name: str = attr.ib()
    ode: ODE = attr.ib()
    function_name: str = attr.ib()
    description: str = attr.ib()
    params: dict[str, typing.Any] | None = attr.ib(default=None)


def rhs_expressions(ode: ODE, function_name: str = "rhs", result_name: str = "dy", params=None):
    """Generate the right hand side of the ODE

    Parameters
    ----------
    ode : ODE
        The ODE
    function_name : str, optional
        The name of the function, by default "rhs"
    result_name : str, optional
        The name of the result, by default "dy"
    params : dict[str, typing.Any], optional
        Additional parameters to pass to the template, by default None

    Returns
    -------
    CodeComponent
        The code component
    """

    descr = f"Compute the right hand side of the {ode} ODE"
    return CodeComponent(
        name="RHSComponent",
        ode=ode,
        function_name=function_name,
        description=descr,
        params=params,
        **{result_name: ode.state_derivatives},
    )



from __future__ import annotations

import abc
import typing
from enum import Enum

import sympy
from sympy.codegen.ast import Assignment
from sympy.printing.codeprinter import CodePrinter

from .. import templates
from ..ode import ODE
from .. import atoms
from .. import schemes


class Func(typing.NamedTuple):
    arguments: list[str]
    states: sympy.IndexedBase
    parameters: sympy.IndexedBase
    values: sympy.IndexedBase
    return_name: str = "values"
    num_return_values: int = 0


class RHSArgument(str, Enum):
    stp = "stp"
    spt = "spt"
    tsp = "tsp"
    tps = "tps"
    pst = "pst"
    pts = "pts"

    @staticmethod
    def get_value(order: str | RHSArgument) -> str:
        if isinstance(order, RHSArgument):
            return order.value
        return str(order)


class SchemeArgument(str, Enum):
    stpd = "stpd"
    sptd = "sptd"
    tspd = "tspd"
    tpsd = "tpsd"
    pstd = "pstd"
    ptsd = "ptsd"
    stdp = "stdp"
    spdt = "spdt"
    tsdp = "tsdp"
    tpds = "tpds"
    psdt = "psdt"
    ptds = "ptds"
    sdtp = "sdtp"
    sdpt = "sdpt"
    tdsp = "tdsp"
    tdps = "tdps"
    pdst = "pdst"
    pdts = "pdts"
    dstp = "dstp"
    dspt = "dspt"
    dtsp = "dtsp"
    dtps = "dtps"
    dpst = "dpst"
    dpts = "dpts"

    @staticmethod
    def get_value(order: str | SchemeArgument) -> str:
        if isinstance(order, SchemeArgument):
            return order.value
        return str(order)


class CodeGenerator(abc.ABC):
    variable_prefix = ""

    def __init__(self, ode: ODE) -> None:
        self.ode = ode

    def _formatter(self, code: str) -> str:
        """Alternative formatter that takes a code snippet
        and output a formatted code snipped together with
        a potential list of errors

        Parameters
        ----------
        code : str
            The code snippet

        Returns
        -------
        str
            formatted code snippet
        """
        return code

    def _format(self, code: str) -> str:
        try:
            formatted_code = self._formatter(code)
        except Exception:
            # FIXME: handle this
            print("An exception was raised")
            formatted_code = code

        return formatted_code

    def _doprint(self, lhs, rhs, use_variable_prefix: bool = False) -> str:
        if use_variable_prefix:
            return f"{self.variable_prefix}{self.printer.doprint(Assignment(lhs, rhs))}"
        return self.printer.doprint(Assignment(lhs, rhs))

    def state_index(self) -> str:
        code = self.template.state_index(
            data={s.name: i for i, s in enumerate(self.ode.sorted_states())}
        )
        return self._format(code)

    def parameter_index(self) -> str:
        code = self.template.parameter_index(
            data={s.name: i for i, s in enumerate(self.ode.parameters)}
        )
        return self._format(code)

    def initial_state_values(self, name="states") -> str:
        """Generate code for initializing state values

        Parameters
        ----------
        name : str, optional
            The name of the variable, by default "states"

        Returns
        -------
        str
            The generated code
        """
        state_result = sympy.IndexedBase(name, shape=(self.ode.num_states,))
        expr = "\n".join(
            [
                self._doprint(state_result[i], value)
                for i, value in enumerate([s.value for s in self.ode.sorted_states()])
            ]
        )

        code = self.template.init_state_values(
            code=expr,
            state_names=[s.name for s in self.ode.sorted_states()],
            state_values=[self.printer.doprint(s.value) for s in self.ode.sorted_states()],
            name=name,
        )
        return self._format(code)

    def initial_parameter_values(self, name="parameters") -> str:
        """Generate code for initializing parameter values

        Parameters
        ----------
        name : str, optional
            The name of the variable, by default "parameters"

        Returns
        -------
        str
            The generated code
        """
        parameter_result = sympy.IndexedBase(name, shape=(self.ode.num_parameters,))

        expr = "\n".join(
            [
                self._doprint(parameter_result[i], value)
                for i, value in enumerate([p.value for p in self.ode.parameters])
            ]
        )
        code = self.template.init_parameter_values(
            code=expr,
            parameter_names=[s.name for s in self.ode.parameters],
            parameter_values=[self.printer.doprint(s.value) for s in self.ode.parameters],
            name=name,
        )
        return self._format(code)

    def _state_assignments(self, states: sympy.IndexedBase) -> str:
        return "\n".join(
            self._doprint(state.symbol, states[i], use_variable_prefix=True)
            for i, state in enumerate(self.ode.sorted_states())
        )

    def _parameter_assignments(self, parameters: sympy.IndexedBase) -> str:
        return "\n".join(
            self._doprint(param.symbol, parameters[i], use_variable_prefix=True)
            for i, param in enumerate(self.ode.parameters)
        )

    def rhs(self, order: RHSArgument | str = RHSArgument.tsp, use_cse=False) -> str:
        """Generate code for the right hand side of the ODE

        Parameters
        ----------
        order : RHSArgument | str, optional
            The order of the arguments, by default RHSArgument.tsp
        use_cse : bool, optional
            Use common subexpression elimination, by default False

        Returns
        -------
        str
            The generated code
        """

        rhs = self._rhs_arguments(order)
        states = self._state_assignments(rhs.states)
        parameters = self._parameter_assignments(rhs.parameters)

        values_lst = []
        index = 0
        values_idx = sympy.IndexedBase("values", shape=(len(self.ode.state_derivatives),))

        for x in self.ode.sorted_assignments():
            values_lst.append(self._doprint(x.symbol, x.expr, use_variable_prefix=True))
            if isinstance(x, atoms.StateDerivative):
                values_lst.append(self._doprint(values_idx[index], x.symbol))
                index += 1

        values = "\n".join(values_lst)
        code = self.template.method(
            name="rhs",
            args=", ".join(rhs.arguments),
            states=states,
            parameters=parameters,
            values=values,
            return_name=rhs.return_name,
            num_return_values=rhs.num_return_values,
        )

        return self._format(code)

    def scheme(self, name: str, order=SchemeArgument.stdp) -> str:
        """Generate code for the forward explicit Euler method

        Parameters
        ----------
        name : str
            The name of the scheme
        order : SchemeArgument | str, optional
            The order of the arguments, by default SchemeArgument.stdp

        Returns
        -------
        str
            The generated code
        """

        rhs = self._scheme_arguments(order)
        states = self._state_assignments(rhs.states)
        parameters = self._parameter_assignments(rhs.parameters)

        dt = sympy.Symbol("dt")
        f = schemes.get_scheme(name)
        eqs = f(self.ode, dt, name=rhs.return_name, printer=self._doprint)
        values = "\n".join(eqs)

        code = self.template.method(
            name=name,
            args=", ".join(rhs.arguments),
            states=states,
            parameters=parameters,
            values=values,
            return_name=rhs.return_name,
            num_return_values=rhs.num_return_values,
        )
        return self._format(code)

    @property
    @abc.abstractmethod
    def printer(self) -> CodePrinter:
        ...

    @property
    @abc.abstractmethod
    def template(self) -> templates.Template:
        ...

    @abc.abstractmethod
    def _rhs_arguments(self, order: RHSArgument | str) -> Func:
        ...

    @abc.abstractmethod
    def _scheme_arguments(self, order: SchemeArgument | str) -> Func:
        ...



from __future__ import annotations
import typing
import sympy
from enum import Enum

from structlog import get_logger

from . import atoms
from .ode import ODE
from . import sympytools

logger = get_logger()


def default_printer(
    lhs: sympy.Symbol | sympy.IndexedBase,
    rhs: sympy.Expr,
    use_variable_prefix: bool = False,
) -> str:
    from sympy.codegen.ast import Assignment
    from sympy.printing import pycode

    return pycode(Assignment(lhs, rhs))


class printer_func(typing.Protocol):
    def __call__(
        self,
        lhs: sympy.Symbol | sympy.IndexedBase,
        rhs: sympy.Expr,
        use_variable_prefix: bool = False,
    ) -> str:
        ...


class scheme_func(typing.Protocol):
    def __call__(
        self,
        ode: ODE,
        dt: sympy.Symbol,
        name: str = "values",
        printer: printer_func = default_printer,
    ) -> list[str]:
        ...


class Scheme(str, Enum):
    forward_explicit_euler = "forward_explicit_euler"
    forward_generalized_rush_larsen = "forward_generalized_rush_larsen"


def get_scheme(scheme: str) -> scheme_func:
    """Get the scheme function from a string"""
    if scheme in ["forward_euler", "forward_explicit_euler", "euler", "explicit_euler"]:
        return forward_explicit_euler
    elif scheme in ["forward_generalized_rush_larsen", "generalized_rush_larsen"]:
        return forward_generalized_rush_larsen
    else:
        raise ValueError(f"Unknown scheme {scheme}")


def fraction_numerator_is_nonzero(expr):
    """Perform a very cheap check to detect if a fraction is definitely non-zero."""

    if isinstance(expr, sympy.Pow):
        # check if the expression is on the form a**-1
        a, b = expr.args
        if b is sympy.S.NegativeOne:
            return True
        else:
            # we won't do any further checks
            return False
    elif isinstance(expr, sympy.Mul):
        # check if all factors are non-zero
        args = expr.args
        certainly_nonzero_args = []
        potentially_nonzero_args = []
        for e in args:
            if len(e.free_symbols) == 0 and e.is_nonzero:
                certainly_nonzero_args.append(e)
            else:
                potentially_nonzero_args.append(e)

        if len(potentially_nonzero_args) == 0:
            # all factors are certainly nonzero
            return True

        # check all potentially non-zero factors
        for e in potentially_nonzero_args:
            if not fraction_numerator_is_nonzero(e):
                return False
        else:
            return True
    else:
        return False


def forward_explicit_euler(
    ode: ODE,
    dt: sympy.Symbol,
    name: str = "values",
    printer: printer_func = default_printer,
) -> list[str]:
    """Generate forward Euler equations for the ODE"""
    eqs = []
    values = sympy.IndexedBase(name, shape=(len(ode.state_derivatives),))
    i = 0
    for x in ode.sorted_assignments():
        eqs.append(printer(x.symbol, x.expr, use_variable_prefix=True))
        if isinstance(x, atoms.StateDerivative):
            eqs.append(
                printer(
                    values[i],
                    x.state.symbol + dt * x.symbol,
                )
            )

            i += 1

    return eqs


def forward_generalized_rush_larsen(
    ode: ODE,
    dt: sympy.Symbol,
    name: str = "values",
    printer: printer_func = default_printer,
    delta=1e-8,
) -> list[str]:
    """Generate forward Generalized Rush Larsen equations for the ODE"""
    eqs = []
    values = sympy.IndexedBase(name, shape=(len(ode.state_derivatives),))
    i = 0
    for x in ode.sorted_assignments():
        eqs.append(printer(x.symbol, x.expr, use_variable_prefix=True))

        if not isinstance(x, atoms.StateDerivative):
            continue

        expr_diff = x.expr.diff(x.state.symbol)

        if expr_diff.is_zero:
            # Use forward Euler
            eqs.append(
                printer(
                    values[i],
                    x.state.symbol + dt * x.symbol,
                )
            )
            i += 1
            continue

        linearized_name = x.name + "_linearized"
        linearized = sympy.Symbol(linearized_name)
        eqs.append(printer(linearized, expr_diff, use_variable_prefix=True))

        need_zero_div_check = not fraction_numerator_is_nonzero(expr_diff)
        if not need_zero_div_check:
            logger.debug(f"{linearized_name} cannot be zero. Skipping zero division check")

        RL_term = x.symbol / linearized * (sympy.exp(linearized * dt) - 1)
        if need_zero_div_check:
            RL_term = sympytools.Conditional(
                abs(linearized) > delta,
                RL_term,
                dt * x.symbol,
            )
        eqs.append(
            printer(
                values[i],
                x.state.symbol + RL_term,
            )
        )
        i += 1
    return eqs



from pathlib import Path
import pytest
import gotranx

here = Path(__file__).parent.absolute()


@pytest.mark.parametrize("grouping", ["encapsulation", "containment"])
@pytest.mark.parametrize(
    "cellml_file, num_states, num_parameters",
    [("noble_1962.cellml", 4, 5), ("ToRORd_dynCl_mid.cellml", 45, 112)],
)
def test_cellml_to_gotran(grouping, cellml_file, num_states, num_parameters):
    params = {"grouping": grouping}
    text = gotranx.cellml.cellml_to_gotran(
        filename=here / "cellml_files" / cellml_file, params=params
    )
    ode = gotranx.load.ode_from_string(text)
    assert ode.num_states == num_states
    assert ode.num_parameters == num_parameters



import typing
from pathlib import Path

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore

import typer
from ..schemes import Scheme


app = typer.Typer()


def version_callback(show_version: bool):
    """Prints version information."""
    if show_version:
        from .. import __version__, __program_name__

        typer.echo(f"{__program_name__} {__version__}")
        raise typer.Exit()


def license_callback(show_license: bool):
    """Prints license information."""
    if show_license:
        from .. import __license__

        typer.echo(f"{__license__}")
        raise typer.Exit()


@app.command()
def main(
    fname: typing.Optional[Path] = typer.Argument(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    to: str = typer.Option(
        "",
        "--to",
        help="Generate code to another programming language",
    ),
    outname: typing.Optional[str] = typer.Option(
        None,
        "-o",
        "--outname",
        help="Output name",
    ),
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show version",
    ),
    license: bool = typer.Option(
        None,
        "--license",
        callback=license_callback,
        is_eager=True,
        help="Show license",
    ),
    scheme: Annotated[typing.Optional[typing.List[Scheme]], typer.Option()] = None,
):
    if fname is None:
        return typer.echo("No file specified")

    if to == "":
        # Check if outname is specified
        if outname is None:
            return typer.echo("No output name specified")
        else:
            to = Path(outname).suffix

    if to in {".c", ".h"}:
        from . import gotran2c

        gotran2c.main(fname=fname, suffix=to, outname=outname, scheme=scheme)
    if to in {".py"}:
        from . import gotran2py

        gotran2py.main(fname=fname, suffix=to, outname=outname, scheme=scheme)

    if to in {".ode"}:
        from .cellml2ode import main as cellml2ode

        cellml2ode(fname=fname, outname=outname)



from __future__ import annotations
from collections import defaultdict

from functools import cached_property
from pathlib import Path
from graphlib import TopologicalSorter
from typing import Iterable
from typing import Sequence
from typing import TypeVar
from typing import cast
from typing import Any

import sympy as sp

from . import atoms
from . import exceptions
from .ode_component import Component

T = TypeVar("T")
U = TypeVar("U", bound=atoms.Assignment)


def check_components(components: Sequence[Component]):
    for comp in components:
        if not comp.is_complete():
            raise exceptions.ComponentNotCompleteError(
                component_name=comp.name,
                missing_state_derivatives=[state.name for state in comp.states_without_derivatives],
            )


def gather_atoms(
    components: Sequence[Component],
) -> tuple[list[str], dict[str, set[Any]], dict[str, sp.Symbol], dict[str, atoms.Atom]]:
    symbol_names = []
    symbols = {}
    symbol_values = defaultdict(set)
    lookup: dict[str, atoms.Atom] = {}
    for component in components:
        for p in component.parameters:
            symbol_names.append(p.name)
            symbols[p.name] = p.symbol
            lookup[p.name] = p
            symbol_values[p.name].add(p.value)
        for s in component.states:
            symbol_names.append(s.name)
            symbols[s.name] = s.symbol
            lookup[s.name] = s
            symbol_values[s.name].add(s.value)
        for i in component.intermediates:
            symbol_names.append(i.name)
            symbols[i.name] = i.symbol
            lookup[i.name] = i
            symbol_values[i.name].add(i.expr)
        for st in component.state_derivatives:
            symbol_names.append(st.name)
            symbols[st.name] = st.symbol
            lookup[st.name] = st
    return symbol_names, symbol_values, symbols, lookup


def find_duplicates(x: Iterable[T]) -> set[T]:
    """Find duplicates in an iterable.
    Assumes type is hashable

    Parameters
    ----------
    x : Iterable[T]
        The iterable with potential duplicates

    Returns
    -------
    set[T]
        List of duplicate values
    """
    seen = set()
    dupes = []

    for xi in x:
        if xi in seen:
            dupes.append(xi)
        else:
            seen.add(xi)
    return set(dupes)


def add_temporal_state(
    components: Sequence[Component],
    t: sp.Symbol,
) -> tuple[Component, ...]:
    new_components = []
    for component in components:
        state_derivatives = set()
        states = set()
        for state_derivative in component.state_derivatives:
            new_state = state_derivative.state.to_TimeDependentState(t)
            state_derivatives.add(
                atoms.StateDerivative(
                    name=state_derivative.name,
                    value=state_derivative.value,
                    components=state_derivative.components,
                    description=state_derivative.description,
                    unit_str=state_derivative.unit_str,
                    unit=state_derivative.unit,
                    state=new_state,
                    symbol=sp.Derivative(new_state.symbol, t),
                ),
            )
            states.add(new_state)

        new_components.append(
            Component(
                name=component.name,
                states=frozenset(states),
                parameters=component.parameters,
                assignments=frozenset(component.intermediates | state_derivatives),
            ),
        )
    return tuple(new_components)


def resolve_expressions(
    components: Sequence[Component],
    symbols: dict[str, sp.Symbol],
) -> tuple[Component, ...]:
    new_components = []
    for component in components:
        assignments = []
        for assignment in component.assignments:
            assignments.append(assignment.resolve_expression(symbols))
        new_components.append(
            Component(
                name=component.name,
                states=component.states,
                parameters=component.parameters,
                assignments=frozenset(assignments),
            ),
        )
    return tuple(new_components)


def make_ode(
    components: Sequence[Component],
    comments: Sequence[atoms.Comment] | None = None,
    name: str = "ODE",
) -> ODE:
    check_components(components=components)
    t = sp.Symbol("t")
    # components = add_temporal_state(components, t)
    check_components(components=components)
    _, symbol_values, symbols, lookup = gather_atoms(components=components)
    symbols["time"] = t
    symbols["t"] = t

    if any(x > 1 for x in map(len, symbol_values.values())):
        raise exceptions.DuplicateSymbolError(
            set(k for k, v in symbol_values.items() if len(v) > 1)
        )
    components = resolve_expressions(components=components, symbols=symbols)
    return ODE(components=components, t=t, name=name, comments=comments)


def sort_assignments(
    assignments: Iterable[atoms.Assignment],
    assignments_only: bool = True,
) -> tuple[str, ...]:
    sorter: TopologicalSorter = TopologicalSorter()
    assignment_names = set()
    for assignment in assignments:
        assignment_names.add(assignment.name)
        sorter.add(assignment.name, *assignment.value.dependencies)

    static_order = tuple(sorter.static_order())

    if assignments_only:
        static_order = tuple([name for name in static_order if name in assignment_names])

    return static_order


class ODE:
    def __init__(
        self,
        components: Sequence[Component],
        t: sp.Symbol | None = None,
        name: str = "ODE",
        comments: Sequence[atoms.Comment] | None = None,
    ):
        check_components(components)
        _, symbol_values, symbols, lookup = gather_atoms(components=components)

        if any(x > 1 for x in map(len, symbol_values.values())):
            raise exceptions.DuplicateSymbolError(
                set(k for k, v in symbol_values.items() if len(v) > 1)
            )

        t = sp.Symbol("t")
        self.t = t
        symbols["time"] = t

        self._symbols = symbols
        self._lookup = lookup
        self.components = components
        self.name = name
        if comments is None:
            comments = (atoms.Comment(""),)

        self.comments = comments
        self.text = "".join(comment.text for comment in comments)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.name}, "
            f"num_states={self.num_states}, "
            f"num_parameters={self.num_parameters})"
        )

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, ODE):
            return False

        return (
            __o.comments == self.comments
            and __o.components == self.components
            and __o.name == self.name
        )

    def simplify(self) -> ODE:
        """Run sympy's simplify function on all expressions in the ODE"""
        return ODE(
            components=tuple([comp.simplify() for comp in self.components]),
            t=self.t,
            name=self.name,
            comments=self.comments,
        )

    @property
    def states(self) -> tuple[atoms.State, ...]:
        states: set[atoms.State] = set()
        for component in self.components:
            states |= component.states
        return tuple(sorted(states, key=lambda x: x.name))

    @property
    def num_states(self) -> int:
        return len(self.states)

    @property
    def num_components(self) -> int:
        return len(self.components)

    @cached_property
    def parameters(self) -> tuple[atoms.Parameter, ...]:
        parameters: set[atoms.Parameter] = set()
        for component in self.components:
            parameters |= component.parameters
        return tuple(sorted(parameters, key=lambda x: x.name))

    @property
    def num_parameters(self) -> int:
        return len(self.parameters)

    @cached_property
    def state_derivatives(self) -> tuple[atoms.StateDerivative, ...]:
        state_derivatives: set[atoms.StateDerivative] = set()
        for component in self.components:
            state_derivatives |= component.state_derivatives
        return tuple(sorted(state_derivatives, key=lambda x: x.name))

    @cached_property
    def intermediates(self) -> tuple[atoms.Intermediate, ...]:
        intermediates: set[atoms.Intermediate] = set()
        for component in self.components:
            intermediates |= component.intermediates
        return tuple(sorted(intermediates, key=lambda x: x.name))

    @property
    def symbols(self) -> dict[str, sp.Symbol]:
        return self._symbols

    def __getitem__(self, name) -> atoms.Atom:
        return self._lookup[name]

    def sorted_assignments(self, assignments_only: bool = True) -> tuple[atoms.Assignment, ...]:
        names = sort_assignments(
            self.intermediates + self.state_derivatives,
            assignments_only=assignments_only,
        )
        return tuple([cast(atoms.Assignment, self[name]) for name in names])

    def sorted_state_derivatives(self) -> tuple[atoms.StateDerivative, ...]:
        return tuple(s for s in self.sorted_assignments() if isinstance(s, atoms.StateDerivative))

    def sorted_states(self) -> tuple[atoms.State, ...]:
        return tuple(s.state for s in self.sorted_state_derivatives())

    def save(self, path: Path) -> None:
        from .save import write_ODE_to_ode_file

        write_ODE_to_ode_file(self, path)



from __future__ import annotations
from pathlib import Path
from structlog import get_logger

from ..cellml import cellml_to_gotran

logger = get_logger()


def main(
    fname: Path,
    outname: str | None = None,
) -> None:
    assert fname.suffix == ".cellml", f"File {fname} must be a cellml file"
    assert fname.exists(), f"File {fname} does not exist"
    logger.info(f"Converting {fname} to gotran ode file")
    code = cellml_to_gotran(fname)
    out = fname.with_suffix(".ode") if outname is None else Path(outname)
    out.write_text(code)
    logger.info(f"Wrote {out}")



from __future__ import annotations

import re

import attr

from . import atoms
from . import exceptions


STATE_DERIV_EXPR = re.compile(r"^d(?P<state>\w+)_dt$")


@attr.s(frozen=True, slots=True)
class Component:
    """A Component is a collection of states, parameters and
    assignments that are given the same tag. This is useful if
    you have a complicated set of equations and need to group them together.

    Raises
    ------
    StateNotFound
        If component contains a state derivative but not the State
    """

    name: str = attr.ib()
    states: frozenset[atoms.State] = attr.ib()
    parameters: frozenset[atoms.Parameter] = attr.ib()
    assignments: frozenset[atoms.Assignment] = attr.ib()
    state_derivatives: frozenset[atoms.StateDerivative] = attr.ib(init=False)
    intermediates: frozenset[atoms.Intermediate] = attr.ib(init=False)

    def __attrs_post_init__(self):
        self._handle_assignments()

    def simplify(self) -> Component:
        """Run sympy's simplify function on all expressions in the ODE"""
        return Component(
            name=self.name,
            states=self.states,
            parameters=self.parameters,
            assignments=frozenset([assignment.simplify() for assignment in self.assignments]),
        )

    def find_state(self, state_name: str) -> atoms.State:
        """Find a state by name

        Parameters
        ----------
        state_name : str
            The name of the state

        Returns
        -------
        atoms.State
            The state

        Raises
        ------
        exceptions.StateNotFoundInComponent
            If state is not found in component
        """
        for state in self.states:
            if state.name == state_name:
                return state
        else:
            raise exceptions.StateNotFoundInComponent(
                state_name=state_name,
                component_name=self.name,
            )

    def find_parameter(self, parameter_name: str) -> atoms.Parameter:
        """Find a parameter by name

        Parameters
        ----------
        parameter_name : str
            The name of the parameter

        Returns
        -------
        atoms.Parameter
            The parameter

        Raises
        ------
        exceptions.ParameterNotFoundInComponent
            If parameter is not found in component
        """

        for parameter in self.parameters:
            if parameter.name == parameter_name:
                return parameter
        else:
            raise exceptions.ParameterNotFoundInComponent(
                parameter_name=parameter_name,
                component_name=self.name,
            )

    def _handle_assignments(self):
        state_derivatives = set()
        intermediates = set()
        for assignment in self.assignments:
            if isinstance(assignment, atoms.Intermediate):
                intermediates.add(assignment)
            elif isinstance(assignment, atoms.StateDerivative):
                state_derivatives.add(assignment)
            else:
                if state_name := STATE_DERIV_EXPR.match(assignment.name):
                    state = self.find_state(state_name=state_name.groupdict()["state"])
                    state_derivatives.add(assignment.to_state_derivative(state))
                else:
                    intermediates.add(assignment.to_intermediate())

        # https://docs.python.org/3/library/dataclasses.html#frozen-instances
        object.__setattr__(self, "intermediates", intermediates)
        object.__setattr__(self, "state_derivatives", state_derivatives)

    def is_complete(self) -> bool:
        """Returns true if all states have a corresponding state derivative"""

        return self.states_with_derivatives == self.states

    @property
    def states_with_derivatives(self) -> frozenset[atoms.State]:
        """Returns the states that have a corresponding state derivative

        Returns
        -------
        frozenset[atoms.State]
            The states that have a corresponding state derivative
        """
        states = set()
        for state_derivative in self.state_derivatives:
            states.add(state_derivative.state)
        return frozenset(states)

    @property
    def states_without_derivatives(self) -> frozenset[atoms.State]:
        """Returns the states that do not have a corresponding state derivative

        Returns
        -------
        frozenset[atoms.State]
            The states that do not have a corresponding state derivative
        """
        return self.states.difference(self.states_with_derivatives)

    @property
    def atoms(self) -> frozenset[atoms.Atom]:
        """Returns all atoms in the component

        Returns
        -------
        frozenset[atoms.Atom]
            All atoms in the component
        """
        return frozenset(
            self.states.union(self.parameters).union(self.assignments).union(self.intermediates)
        )



import pytest
import sympy
from gotranx.ode import make_ode
from gotranx.ode import ODE
from gotranx import sympytools


@pytest.fixture(scope="module")
def ode(trans, parser) -> ODE:
    expr = """
    parameters(a=0)
    parameters("My component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states("My component", x=1.0, z=3.05, y=2.0)

    expressions("My component")
    dx_dt = sigma*(-x + y)
    dy_dt = y_int - y # millivolt
    dz_dt = z_int + x*y
    y_int = x*(rho - z)
    z_int = -beta*z
    """
    tree = parser.parse(expr)
    return make_ode(*trans.transform(tree))


# @pytest.fixture(scope="module")
# def sym_ode(ode: ODE) -> SympyODE:
#     return SympyODE(ode)


def test_states(ode):
    states = sympytools.states_matrix(ode)
    assert len(states) == 3
    assert str(states[0]) == "x"
    assert str(states[1]) == "y"
    assert str(states[2]) == "z"


def test_Conditional_boolean_condition():
    assert sympytools.Conditional(True, 1, 2) == 1


def test_Conditional_invalid_condition():
    with pytest.raises(TypeError):
        sympytools.Conditional(sympy.Eq, 1, 2)


def test_ContinousConditional_invalid_condition():
    with pytest.raises(TypeError):
        sympytools.ContinuousConditional(sympy.Eq, 1, 2)


def test_rhs_matrx(ode: ODE):
    rhs = sympytools.rhs_matrix(ode)
    assert len(rhs) == 3
    assert str(rhs[0]) == "sigma*(-x + y)"
    assert str(rhs[1]) == "x*(rho - z) - y"
    assert str(rhs[2]) == "(-beta)*z + x*y"


def test_jacobian_matrix(ode: ODE):
    jac = sympytools.jacobi_matrix(ode)

    assert str(jac[0]) == "-sigma"
    assert str(jac[1]) == "sigma"
    assert str(jac[2]) == "0"

    assert str(jac[3]) == "rho - z"
    assert str(jac[4]) == "-1"
    assert str(jac[5]) == "-x"

    assert str(jac[6]) == "y"
    assert str(jac[7]) == "x"
    assert str(jac[8]) == "-beta"



import lark
import pytest
import sympy as sp
from gotranx import atoms


@pytest.mark.parametrize(
    "expr, deps",
    [
        ("x = (1 * y) + rho - (z / sigma)", {"y", "rho", "z", "sigma"}),
        ("x = (1 * y) + rho - (y / 2)", {"y", "rho"}),
        ("x = 1", set()),
    ],
)
def test_expression_dependencies(expr, deps, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert result[0].value.dependencies == deps


def test_parameter_arguments(parser, trans):
    expr = """parameters("My component",
        y = 1.0,
        x = ScalarParam(42.0),
        w = ScalarParam(10.2, unit="mM"),
        v = ScalarParam(3.14, unit="mV", description="Description of v"),
        z = ScalarParam(42.0, description="Description of z")
    )
    """

    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert result[0] == atoms.Parameter(
        name="y",
        value=sp.sympify(1.0),
        components=("My component",),
    )
    assert result[1] == atoms.Parameter(
        name="x",
        value=sp.sympify(42.0),
        components=("My component",),
    )
    assert result[2] == atoms.Parameter(
        name="w",
        value=sp.sympify(10.2),
        components=("My component",),
        unit_str="mM",
    )
    assert result[3] == atoms.Parameter(
        name="v",
        value=sp.sympify(3.14),
        description="Description of v",
        components=("My component",),
        unit_str="mV",
    )
    assert result[4] == atoms.Parameter(
        name="z",
        value=sp.sympify(42.0),
        description="Description of z",
        components=("My component",),
    )


def test_states_arguments(parser, trans):
    expr = """
    states("My component",
        y = 1.0,
        x = ScalarParam(42.0)
    )
    states("My component", "Some info",
        w = ScalarParam(10.2, unit="mM"),
        v = ScalarParam(3.14, unit="mV", description="Description of v")
    )
    states("My other component", "other info",
        z = ScalarParam(42.0, description="Description of z")
    )
    """

    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result.components) == 4

    comp1 = result.components[0]
    assert comp1.name == "My component"

    assert comp1.states == {
        atoms.State(name="y", value=sp.sympify(1.0), components=("My component",)),
        atoms.State(name="x", value=sp.sympify(42.0), components=("My component",)),
        atoms.State(
            name="w",
            value=sp.sympify(10.2),
            components=("My component", "Some info"),
            unit_str="mM",
        ),
        atoms.State(
            name="v",
            value=sp.sympify(3.14),
            description="Description of v",
            components=("My component", "Some info"),
            unit_str="mV",
        ),
    }

    comp2 = result.components[2]
    assert comp2.name == "My other component"
    assert comp2.states == {
        atoms.State(
            name="z",
            value=sp.sympify(42.0),
            description="Description of z",
            components=("My other component", "other info"),
        ),
    }


def test_comment(parser, trans):
    expr = """
    # This is one comment.
    # Here is another comment
    x = 1
    parameters(y = 2)
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)

    assert len(result.comments) == 1
    assert result.comments[0] == atoms.Comment(
        text="This is one comment. Here is another comment",
    )

    comp = result.components
    assert len(comp) == 1
    assert comp[0].parameters == {atoms.Parameter(name="y", value=sp.sympify(2))}

    assert comp[0].assignments == {
        atoms.Assignment(
            name="x",
            value=atoms.Expression(
                tree=lark.Tree("scientific", [lark.Token("SCIENTIFIC_NUMBER", "1")]),
            ),
        ),
    }


def test_TimeDependentState(parser, trans):
    expr = """
    states(x = ScalarParam(1, unit="M"))
    """
    tree = parser.parse(expr)
    result = trans.transform(tree)
    t = sp.Symbol("t")
    x = result[0]
    x_t = x.to_TimeDependentState(t)

    assert str(x.symbol) == "x"
    assert str(x_t.symbol) == "x(t)"


# def test_StateDerivative(parser, trans):
#     expr = """
#     parameters(a = 1)
#     states(x=2)
#     dx_dt = a + 1
#     """
#     tree = parser.parse(expr)
#     result = trans.transform(tree)
#     assert len(result) == 1
#     comp = result[0]
#     assert len(comp.state_derivatives) == 1
#     dx_dt = comp.state_derivatives.pop()
#     t = sp.Symbol("t")
#     breakpoint()



from __future__ import annotations
import typing

from . import c
from . import python


class Template(typing.Protocol):
    @staticmethod
    def state_index(data: dict[str, float]) -> str:
        ...

    @staticmethod
    def parameter_index(data: dict[str, float]) -> str:
        ...

    @staticmethod
    def init_state_values(
        name: str, state_values: list[float], state_names: list[str], code: str
    ) -> str:
        ...

    @staticmethod
    def init_parameter_values(
        name: str, parameter_values: list[float], parameter_names: list[str], code: str
    ) -> str:
        ...

    @staticmethod
    def method(
        name: str,
        args: str,
        states: str,
        parameters: str,
        values: str,
        return_name: str | None,
        num_return_values: int,
    ) -> str:
        ...


__all__ = ["c", "python", "Template"]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import lark


class GotranParserError(Exception):
    pass


@dataclass
class ODEFileNotFound(GotranParserError):
    fname: Path


@dataclass
class StateNotFoundInComponent(GotranParserError):
    state_name: str
    component_name: str

    def __str__(self) -> str:
        return (
            f"State with name {self.state_name!r} "
            f"not found in component {self.component_name!r}"
        )


@dataclass
class ParameterNotFoundInComponent(GotranParserError):
    parameter_name: str
    component_name: str

    def __str__(self) -> str:
        return (
            f"Parameter with name {self.parameter_name!r} "
            f"not found in component {self.component_name!r}"
        )


@dataclass
class ComponentNotCompleteError(GotranParserError):
    component_name: str
    missing_state_derivatives: list[str]

    def __str__(self) -> str:
        return (
            f"Component {self.component_name!r} is not complete. "
            f"Missing state derivatives for {self.missing_state_derivatives!r}"
        )


@dataclass
class DuplicateSymbolError(GotranParserError):
    duplicates: set[str]

    def __str__(self) -> str:
        return (
            f"Found multiple definitions for {self.duplicates!r}. "
            "Please make sure to only define each variable once in order "
            "to avoid ambiguity. Equations should be reorderable."
        )


@dataclass
class UnknownTreeTypeError(GotranParserError):
    datatype: str
    atom: str

    def __str__(self) -> str:
        return f"Cannot prase tree data type {self.datatype} for atom {self.atom}"


@dataclass
class InvalidTreeError(GotranParserError):
    tree: lark.Tree

    def __str__(self) -> str:
        return f"Invaild tree with data attribute {self.tree.data!r} \n{self.tree.pretty()}"



import pytest
import sympy
from gotranx.ode import make_ode
from gotranx.ode import ODE
from gotranx import schemes


@pytest.fixture(scope="module")
def ode(trans, parser) -> ODE:
    expr = """
    parameters(a=0)
    parameters("My component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states("My component", x=1.0, z=3.05, y=2.0)

    expressions("My component")
    dx_dt = sigma*(-x + y)
    dy_dt = y_int - y # millivolt
    dz_dt = z_int + x*y
    y_int = x*(rho - z)
    z_int = -beta*z
    """
    tree = parser.parse(expr)
    return make_ode(*trans.transform(tree))


def test_forward_explicit_euler(ode: ODE):
    dt = sympy.Symbol("dt")
    eqs = schemes.forward_explicit_euler(ode, dt)
    assert len(eqs) == 8

    assert eqs[0] == "y_int = x*(rho - z)"
    assert eqs[1] == "z_int = (-beta)*z"
    assert eqs[2] == "dx_dt = sigma*(-x + y)"
    assert eqs[3] == "values[0] = dt*dx_dt + x"
    assert eqs[4] == "dy_dt = -y + y_int"
    assert eqs[5] == "values[1] = dt*dy_dt + y"
    assert eqs[6] == "dz_dt = x*y + z_int"
    assert eqs[7] == "values[2] = dt*dz_dt + z"


def test_forward_generalized_rush_larsen(ode: ODE):
    dt = sympy.Symbol("dt")
    eqs = schemes.forward_generalized_rush_larsen(ode, dt)

    assert len(eqs) == 10

    assert str(eqs[0]) == "y_int = x*(rho - z)"
    assert str(eqs[1]) == "z_int = (-beta)*z"
    assert str(eqs[2]) == "dx_dt = sigma*(-x + y)"
    assert str(eqs[3]) == "dx_dt_linearized = -sigma"
    assert str(eqs[4]) == (
        "values[0] = x + "
        "((dx_dt*(math.exp(dt*dx_dt_linearized) - 1)"
        "/dx_dt_linearized) if (abs(dx_dt_linearized) > 1.0e-8) "
        "else (dt*dx_dt))"
    )
    assert str(eqs[5]) == "dy_dt = -y + y_int"
    assert str(eqs[6]) == "dy_dt_linearized = -1"
    assert str(eqs[7]) == (
        "values[1] = y + "
        "((dy_dt*(math.exp(dt*dy_dt_linearized) - 1)"
        "/dy_dt_linearized) if (abs(dy_dt_linearized) > 1.0e-8) "
        "else (dt*dy_dt))"
    )



from pathlib import Path

import pytest
from gotranx import load_ode
from gotranx import save
from gotranx.ode import make_ode


@pytest.fixture
def path():
    path_ = Path("tmp_file.ode")
    yield path_
    path_.unlink()


def test_write_ode_file_write_dummy(path, parser, trans):
    expr = """
    # This is an ODE.
    # Here is another line.
    # And here is a third line, just to make sure
    # that the total line length is more than 80 characters.
    states("First component", "X-gate", x = 1, xr=3.14)
    states("First component", "Y-gate", y = 1)
    states("Second component", z=1)
    parameters("First component", a=1, b=2)
    parameters("Second component", c=3)

    expressions("First component")
    d = a + b * 2 - 3 / c

    expressions("First component", "X-gate")
    dx_dt=a+1
    dxr_dt = (-d) * xr + (x / b)

    expressions("First component", "Y-gate")
    dy_dt = 2 * d - 1

    expressions("Second component")
    dz_dt = 1 + x - y
    """
    tree = parser.parse(expr)
    old_ode = make_ode(*trans.transform(tree), name=path.stem)
    save.write_ODE_to_ode_file(old_ode, path)
    ode = load_ode(path)
    assert ode.simplify() == old_ode.simplify()


def test_write_ode_file_write_lorentz(path, parser, trans):
    expr = """
    parameters(
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states(x=1.0, y=2.0,z=3.05)

    dx_dt = sigma*(-x + y)
    dy_dt = x*(rho - z) - y # millivolt
    dz_dt = -beta*z + x*y
    """
    tree = parser.parse(expr)
    old_ode = make_ode(*trans.transform(tree), name=path.stem)
    save.write_ODE_to_ode_file(old_ode, path)
    ode = load_ode(path)
    assert ode == old_ode



from __future__ import annotations

# Copyright (C) 2013 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["MathMLBaseParser", "MathMLCPPParser"]
import sys
import structlog

logger = structlog.getLogger(__name__)

_all_keywords: list[str] = []


class MathMLBaseParser(object):
    def __init__(self):
        self._state_variable = None
        self._derivative = None
        self.variables_names = set()

        self._precedence = {
            "piecewise": 0,
            "power": 0,
            "rem": 0.5,
            "divide": 1,
            "times": 2,
            "minus": 4,
            "plus": 5,
            "lt": 6,
            "gt": 6,
            "leq": 6,
            "geq": 6,
            "eq": 10,
            "exp": 10,
            "ln": 10,
            "abs": 10,
            "floor": 10,
            "log": 10,
            "root": 10,
            "tan": 10,
            "cos": 10,
            "sin": 10,
            "tanh": 10,
            "cosh": 10,
            "sinh": 10,
            "arccos": 10,
            "arcsin": 10,
            "arctan": 10,
        }

        self._operators = {
            "power": "**",
            "rem": " % ",
            "divide": "/",
            "times": "*",
            "minus": " - ",
            "plus": " + ",
            "lt": " < ",
            "gt": " > ",
            "leq": " <= ",
            "geq": " >= ",
            "eq": " = ",
            "exp": "exp",
            "ln": "log",
            "abs": "abs",
            "floor": "floor",
            "log": "log",
            "root": "sqrt",
            "tan": "tan",
            "cos": "cos",
            "sin": "sin",
            "tanh": "tanh",
            "cosh": "cosh",
            "sinh": "sinh",
            "arccos": "acos",
            "arcsin": "asin",
            "arctan": "atan",
        }

    def use_parenthesis(self, child, parent, first_operand=True):
        """
        Return true if child operation need parenthesis
        """
        if parent is None:
            return False

        parent_prec = self._precedence[parent]
        if parent == "minus" and not first_operand:
            parent_prec -= 0.5

        if parent == "divide" and child == "times" and first_operand:
            return False

        if parent == "minus" and child == "plus" and first_operand:
            return False

        return parent_prec < self._precedence[child]

    def __getitem__(self, operator):
        return self._operators[operator]

    def _gettag(self, node):
        """
        Splits off the namespace part from name, and returns the rest, the tag
        """
        return "".join(node.tag.split("}")[1:])

    def parse(self, root):
        """
        Recursively parse a mathML subtree and return an list of tokens
        together with any state variable and derivative.
        """
        self._state_variable = None
        self._derivative = None
        self.used_variables = set()

        equation_list = self._parse_subtree(root)
        return (
            equation_list,
            self._state_variable,
            self._derivative,
            self.used_variables,
        )

    def _parse_subtree(self, root, parent=None, first_operand=True):
        op = self._gettag(root)

        # If the tag i "apply" pick the operator and continue parsing
        if op == "apply":
            children = list(root)
            op = self._gettag(children[0])
            root = children[1:]
        # If use special method to parse
        if hasattr(self, "_parse_" + op):
            return getattr(self, "_parse_" + op)(root, parent)
        elif op in list(self._operators.keys()):
            # Build the equation string
            eq = []

            # Check if we need parenthesis
            use_parent = self.use_parenthesis(op, parent, first_operand)

            # If unary operator
            if len(root) == 1:
                # Special case if operator is "minus"
                if op == "minus":
                    # If an unary minus is infront of a cn or ci we skip
                    # parenthesize
                    if self._gettag(root[0]) in ["ci", "cn"]:
                        use_parent = False

                    # If an unary minus is infront of a plus we always use parenthesize
                    if self._gettag(root[0]) == "apply" and self._gettag(
                        list(root[0])[0],
                    ) in ["plus", "minus"]:
                        use_parent = True

                    eq += ["-"]

                else:
                    # Always use paranthesis for unary operators
                    use_parent = True
                    eq += [self._operators[op]]

                eq += ["("] * use_parent + self._parse_subtree(root[0], op) + [")"] * use_parent
                return eq
            else:
                # Binary operator
                eq += ["("] * use_parent + self._parse_subtree(root[0], op)
                for operand in root[1:]:
                    eq = (
                        eq
                        + [self._operators[op]]
                        + self._parse_subtree(operand, op, first_operand=False)
                    )
                eq = eq + [")"] * use_parent

                return eq
        else:
            logger.error("No support for parsing MathML " + op + " operator.")

    def _parse_conditional(self, condition, operands, parent):
        return (
            [condition]
            + ["("]
            + self._parse_subtree(operands[0], parent)
            + [", "]
            + self._parse_subtree(operands[1], parent)
            + [")"]
        )

    def _parse_and(self, operands, parent):
        ret = ["And("]
        for operand in operands:
            ret += self._parse_subtree(operand, parent) + [", "]
        return ret + [")"]

    def _parse_or(self, operands, parent):
        ret = ["Or("]
        for operand in operands:
            ret += self._parse_subtree(operand, parent) + [", "]
        return ret + [")"]

    def _parse_lt(self, operands, parent):
        return self._parse_conditional("Lt", operands, "lt")

    def _parse_leq(self, operands, parent):
        return self._parse_conditional("Le", operands, "leq")

    def _parse_gt(self, operands, parent):
        return self._parse_conditional("Gt", operands, "gt")

    def _parse_geq(self, operands, parent):
        return self._parse_conditional("Ge", operands, "geq")

    def _parse_neq(self, operands, parent):
        return self._parse_conditional("Ne", operands, "neq")

    def _parse_eq(self, operands, parent):
        # Parsing conditional
        if parent == "piecewise":
            return self._parse_conditional("Eq", operands, "eq")

        # Parsing assignment
        return (
            self._parse_subtree(operands[0], "eq")
            + [self["eq"]]
            + self._parse_subtree(operands[1], "eq")
        )

    def _parse_pi(self, var, parent):
        return ["pi"]

    def _parse_ci(self, var, parent):
        varname = var.text.strip()
        if varname in _all_keywords:
            varname = varname + "_"
        self.used_variables.add(varname)
        return [varname]

    def _parse_cn(self, var, parent):
        value = var.text.strip()
        if "type" in list(var.keys()) and var.get("type") == "e-notation":
            # Get rid of potential float repr
            exponent = "e" + str(int(list(var)[0].tail.strip()))
        else:
            exponent = ""
        value += exponent

        return [value]

    def _parse_diff(self, operands, parent):
        # Store old used_variables so we can erase any collected state
        # variables
        used_variables_prior_parse = self.used_variables.copy()

        x = "".join(self._parse_subtree(operands[1], "diff"))
        y = "".join(self._parse_subtree(operands[0], "diff"))

        if x in _all_keywords:
            x = x + "_"

        if y == "time":
            y = "t"

        d = "d" + x + "_d" + y

        # Restore used_variables
        self.used_variables = used_variables_prior_parse

        # Store derivative
        self.used_variables.add(d)

        # This is an in/out variable remember it
        self._derivative = d
        self._state_variable = x
        return [d]

    def _parse_bvar(self, var, parent):
        if len(var) == 1:
            return self._parse_subtree(var[0], "bvar")
        else:
            logger.error("ERROR: No support for higher order derivatives.")

    def _parse_piecewise(self, cases, parent=None):
        if len(cases) == 2:
            piece_children = list(cases[0])
            cond = self._parse_subtree(piece_children[1], "piecewise")
            true = self._parse_subtree(piece_children[0])
            false = self._parse_subtree(list(cases[1])[0])
            return ["Conditional", "("] + cond + [", "] + true + [", "] + false + [")"]
        else:
            piece_children = list(cases[0])
            cond = self._parse_subtree(piece_children[1], "piecewise")
            true = self._parse_subtree(piece_children[0])
            return (
                ["Conditional", "("]
                + cond
                + [", "]
                + true
                + [", "]
                + self._parse_piecewise(cases[1:])
                + [")"]
            )


class MathMLCPPParser(MathMLBaseParser):
    def _parse_power(self, operands):
        return (
            ["pow", "("]
            + self._parse_subtree(operands[0])
            + [", "]
            + self._parse_subtree(operands[1])
            + [")"]
        )

    def _parse_piecewise(self, cases):
        if len(cases) == 2:
            piece_children = cases[0].getchildren()
            cond = self._parse_subtree(piece_children[1])
            true = self._parse_subtree(piece_children[0])
            false = self._parse_subtree(cases[1].getchildren()[0])
            return ["("] + cond + ["?"] + true + [":"] + false + [")"]
        else:
            sys.exit(
                "ERROR: No support for cases with other than two " "possibilities.",
            )



import pytest
from gotranx import exceptions
from gotranx import ode


def test_ODE_with_incomplete_component_raises_ComponentNotCompleteError(parser, trans):
    expr = "states(x=1, y=2)\n dy_dt=0"
    tree = parser.parse(expr)
    result = trans.transform(tree)
    with pytest.raises(exceptions.ComponentNotCompleteError) as e:
        ode.ODE(result.components)
    assert str(e.value) == "Component '' is not complete. Missing state derivatives for ['x']"


@pytest.mark.parametrize(
    "input, output",
    [
        ((), set()),
        (("x", "y", "z"), set()),
        (("x", "y", "y"), {"y"}),
        (("x"), set()),
        (("y", "y", "y"), {"y"}),
        (("x", "y", "y", "x"), {"x", "y"}),
    ],
)
def test_find_duplicates(input, output):
    assert ode.find_duplicates(input) == output


def test_ODE_with_duplicates_raises_DuplicateSymbolError(parser, trans):
    expr = "states(y=2)\n dy_dt=0 \n y=42"
    tree = parser.parse(expr)
    with pytest.raises(exceptions.DuplicateSymbolError) as e:
        ode.ODE(trans.transform(tree).components)

    assert "Found multiple definitions for {'y'}" in str(e.value)


def test_make_ode_with_duplicates_raises_DuplicateSymbolError(parser, trans):
    expr = "states(y=2)\n dy_dt=0 \n y=42"
    tree = parser.parse(expr)
    with pytest.raises(exceptions.DuplicateSymbolError) as e:
        ode.make_ode(trans.transform(tree).components)

    assert "Found multiple definitions for {'y'}" in str(e.value)


def test_make_ode(parser, trans):
    expr = """
    # This is an ODE.
    # Here is another line.
    states("First component", "X-gate", x = 1, xr=3.14)
    states("First component", "Y-gate", y = 1)
    states("Second component", z=1)
    parameters("First component", a=1, b=2)
    parameters("Second component", c=3)

    expressions("First component")
    d = a + b * 2 - 3 / c

    expressions("First component", "X-gate")
    dx_dt=a+1
    dxr_dt = (x / b) - d * xr

    expressions("First component", "Y-gate")
    dy_dt = 2 * d - 1

    expressions("Second component")
    dz_dt = 1 + x - y
    """
    tree = parser.parse(expr)

    result = ode.make_ode(*trans.transform(tree), name="TestODE")

    assert result.text == "This is an ODE. Here is another line."
    assert repr(result) == "ODE(TestODE, num_states=4, num_parameters=3)"

    assert result.num_components == 4

    states = result.states
    assert len(states) == result.num_states == 4
    state_names = set()
    state_symbols = set()
    for state in states:
        # assert isinstance(state, atoms.TimeDependentState)
        state_names.add(state.name)
        state_symbols.add(str(state.symbol))

    assert state_names == {"x", "xr", "y", "z"}
    # assert state_symbols == {"x(t)", "xr(t)", "y(t)", "z(t)"}

    parameters = result.parameters
    assert len(parameters) == result.num_parameters == 3
    parameter_names = set()
    for parameter in parameters:
        parameter_names.add(parameter.name)

    assert parameter_names == {"a", "b", "c"}

    state_derivatives = result.state_derivatives
    assert len(state_derivatives) == 4
    state_derivative_names = set()
    state_derivative_symbols = set()
    for state_derivative in state_derivatives:
        # assert isinstance(state_derivative.state, atoms.TimeDependentState)
        state_derivative_names.add(state_derivative.name)
        state_derivative_symbols.add(str(state_derivative.symbol))

    assert state_derivative_names == {"dx_dt", "dxr_dt", "dy_dt", "dz_dt"}
    # assert state_derivative_symbols == {
    #     "Derivative(x(t), t)",
    #     "Derivative(xr(t), t)",
    #     "Derivative(y(t), t)",
    #     "Derivative(z(t), t)",
    # }

    intermediates = result.intermediates
    assert len(intermediates) == 1


@pytest.mark.parametrize(
    "assignments_only, expected",
    [
        (True, ("y", "a", "z", "x", "dV_dt")),
        (False, ("V", "y", "a", "z", "x", "dV_dt")),
    ],
)
def test_sort_assignment(assignments_only, expected, parser, trans):
    expr = """
    states(V=0)
    dV_dt = x * z
    x = y + z + V
    z = a - 2
    a = y + 3
    y = 2 * V
    """
    tree = parser.parse(expr)
    result = ode.make_ode(*trans.transform(tree), name="TestODE")

    sorted_assignments = result.sorted_assignments(
        assignments_only=assignments_only,
    )
    assert tuple([x.name for x in sorted_assignments]) == expected



from __future__ import annotations
from sympy.printing.c import C99CodePrinter
from sympy.codegen.ast import Assignment
import sympy

from ..ode import ODE
from .. import templates
from .base import CodeGenerator, Func, RHSArgument, SchemeArgument


class GotranCCodePrinter(C99CodePrinter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._settings["contract"] = False

    def _print_Float(self, flt):
        return self._print(str(float(flt)))

    def _print_Piecewise(self, expr):
        fst, snd = expr.args
        if isinstance(fst[0], Assignment):
            value = (
                f"{super()._print(fst[0].args[0])} = "
                f"({super()._print(fst[1])}) ? "
                f"{super()._print(fst[0].args[1])} : "
                f"{super()._print(snd[0].args[1])};"
            )
        else:
            value = super()._print_Piecewise(expr)

        return value


class CCodeGenerator(CodeGenerator):
    variable_prefix = "const double "

    def __init__(self, ode: ODE, apply_clang_format: bool = True) -> None:
        super().__init__(ode)
        self._printer = GotranCCodePrinter()

        if apply_clang_format:
            try:
                import clang_format_docs
            except ImportError:
                print("Cannot apply clang-format, please install 'clang-format-docs'")
            else:
                setattr(self, "_formatter", clang_format_docs.clang_format_str)

    @property
    def printer(self):
        return self._printer

    @property
    def template(self):
        return templates.c

    def _rhs_arguments(
        self, order: RHSArgument | str = RHSArgument.stp, const_states: bool = True
    ) -> Func:
        value = RHSArgument.get_value(order)
        states_prefix = "const " if const_states else ""
        argument_dict = {
            "s": states_prefix + "double *__restrict states",
            "t": "const double t",
            "p": "const double *__restrict parameters",
        }
        argument_list = [argument_dict[v] for v in value] + ["double* values"]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
        )

    def _scheme_arguments(
        self,
        order: SchemeArgument | str = SchemeArgument.stdp,
        const_states: bool = True,
    ) -> Func:
        value = SchemeArgument.get_value(order)
        states_prefix = "const " if const_states else ""
        argument_dict = {
            "s": states_prefix + "double *__restrict states",
            "t": "const double t",
            "d": "const double dt",
            "p": "const double *__restrict parameters",
        }
        argument_list = [argument_dict[v] for v in value] + ["double* values"]
        states = sympy.IndexedBase("states", shape=(self.ode.num_states,))
        parameters = sympy.IndexedBase("parameters", shape=(self.ode.num_parameters,))
        values = sympy.IndexedBase("values", shape=(self.ode.num_states,))

        return Func(
            arguments=argument_list,
            states=states,
            parameters=parameters,
            values=values,
        )



from importlib.metadata import metadata

from . import cli
from . import codecomponent
from . import exceptions
from . import load
from . import ode
from . import ode_component
from . import parser
from . import codegen
from . import transformer
from . import units
from . import sympytools
from . import schemes
from . import templates
from . import cellml
from .load import load_ode
from .ode import ODE
from .ode_component import Component
from .parser import Parser
from .transformer import TreeToODE


meta = metadata("gotranx")
__version__ = meta["Version"]
__author__ = meta["Author"]
__license__ = meta["License"]
__email__ = meta["Author-email"]
__program_name__ = meta["Name"]

__all__ = [
    "cli",
    "parser",
    "Parser",
    "codegen",
    "transformer",
    "TreeToODE",
    "exceptions",
    "load",
    "load_ode",
    "ode_component",
    "Component",
    "ode",
    "ODE",
    "units",
    "codecomponent",
    "sympytools",
    "schemes",
    "templates",
    "cellml",
]



import pytest
from gotranx import atoms


@pytest.mark.parametrize(
    "expr",
    [
        "parameters(x=1)",
        "parameters(x = 1)",
        "parameters(x= 1)",
        """parameters(
    x=1)""",
        """parameters(
    x=1
    )""",
    ],
)
def test_parameters_single(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 1
    assert result[0] == atoms.Parameter(name="x", value=1)


@pytest.mark.parametrize(
    "expr",
    [
        "parameters(x=1, y=2)",
        "parameters(x = 1, y =2)",
        "parameters(x= 1,y= 2)",
        """parameters(
    x=1, y=2)""",
        """parameters(
    x=1,
    y=2
    )""",
    ],
)
def test_parameters_double(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 2
    assert result[0] == atoms.Parameter(name="x", value=1)
    assert result[1] == atoms.Parameter(name="y", value=2)


def test_parameters_with_component(parser, trans):
    expr = 'parameters("My Component", x=1, y=2)'
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 2
    assert result[0] == atoms.Parameter(name="x", value=1, components=("My Component",))
    assert result[1] == atoms.Parameter(name="y", value=2, components=("My Component",))


def test_different_sets_of_parameters(parser, trans):
    expr = (
        'parameters("First component", x=1, y=2)\n'
        'parameters("Second component", z=3)\nparameters(w=4)'
    )
    tree = parser.parse(expr)
    result = trans.transform(tree).components

    assert len(result) == 3
    first_component = result[0]
    assert first_component.name == "First component"
    assert first_component.parameters == {
        atoms.Parameter(name="x", value=1, components=("First component",)),
        atoms.Parameter(name="y", value=2, components=("First component",)),
    }

    second_component = result[1]
    assert second_component.name == "Second component"
    assert second_component.parameters == {
        atoms.Parameter(name="z", value=3, components=("Second component",)),
    }

    third_component = result[2]
    assert third_component.name == ""
    assert third_component.parameters == {
        atoms.Parameter(name="w", value=4, components=("",)),
    }


@pytest.mark.parametrize(
    "expr, expected",
    [
        (
            'parameters(x=ScalarParam(1, unit="pA"))',
            (atoms.Parameter(name="x", value=1.0, unit_str="pA"),),
        ),
        (
            'parameters(x=ScalarParam(1, unit="pA", description="Info about x"))',
            (
                atoms.Parameter(
                    name="x",
                    value=1.0,
                    unit_str="pA",
                    description="Info about x",
                ),
            ),
        ),
        (
            'parameters(x=ScalarParam(1, unit="pA*pF**-1", description="Info about x"))',
            (
                atoms.Parameter(
                    name="x",
                    value=1.0,
                    unit_str="pA*pF**-1",
                    description="Info about x",
                ),
            ),
        ),
        (
            """
        parameters(
            x=ScalarParam(1, unit="pA", description="Info about x"),
            y=ScalarParam(2.0, unit="mM", description="Info about y")
        )
        """,
            (
                atoms.Parameter(
                    name="x",
                    value=1.0,
                    unit_str="pA",
                    description="Info about x",
                ),
                atoms.Parameter(
                    name="y",
                    value=2.0,
                    unit_str="mM",
                    description="Info about y",
                ),
            ),
        ),
    ],
)
def test_parameter_with_unit_and_desc(expr, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert result == expected



from __future__ import annotations
from textwrap import dedent, indent
import functools


def acc(all_values: str, next_value: str = "# ") -> str:
    if len(all_values.split("\n")[-1]) + len(next_value) > 60:
        return all_values + "\n#" + next_value
    else:
        return all_values + ", " + next_value


def state_index(data: dict[str, int]) -> str:
    return dedent(
        f'''
def state_index(name: str) -> float:
    """Return the index of the state with the given name

    Arguments
    ---------
    name : str
        The name of the state

    Returns
    -------
    float
        The index of the state

    Raises
    ------
    KeyError
        If the name is not a valid state
    """

    data = {repr(data)}
    return data[name]
''',
    )


def init_state_values(name, state_names, state_values, code):
    values_comment = indent(
        "#" + functools.reduce(acc, [f"{n}={v}" for n, v in zip(state_names, state_values)]),
        "    ",
    )

    values = ", ".join(map(str, state_values))
    return dedent(
        f'''
def init_state_values(**values):
    """Initialize state values
    """
{values_comment}

    {name} = numpy.array([{values}])

    for key, value in values.items():
        {name}[state_index(key)] = value

    return {name}
''',
    )


def parameter_index(data: dict[str, int]) -> str:
    return dedent(
        f'''
def parameter_index(name: str) -> float:
    """Return the index of the parameter with the given name

    Arguments
    ---------
    name : str
        The name of the parameter

    Returns
    -------
    float
        The index of the parameter

    Raises
    ------
    KeyError
        If the name is not a valid parameter
    """

    data = {repr(data)}
    return data[name]
''',
    )


def init_parameter_values(name, parameter_names, parameter_values, code):
    values_comment = indent(
        "#"
        + functools.reduce(acc, [f"{n}={v}" for n, v in zip(parameter_names, parameter_values)]),
        "    ",
    )

    values = ", ".join(map(str, parameter_values))
    return dedent(
        f'''
def init_parameter_values(**values):
    """Initialize parameter values
    """
{values_comment}

    {name} = numpy.array([{values}])

    for key, value in values.items():
        {name}[parameter_index(key)] = value

    return {name}
''',
    )


def method(name, args, states, parameters, values, return_name: str, num_return_values: int):
    indent_states = indent(states, "    ")
    indent_parameters = indent(parameters, "    ")
    indent_values = indent(values, "    ")
    indent_return = indent("return " + return_name, "    ")
    return dedent(
        f"""
def {name}({args}):

    # Assign states
{indent_states}

    # Assign parameters
{indent_parameters}

    # Assign expressions
    {return_name} = numpy.zeros({num_return_values})
{indent_values}

{indent_return}
""",
    )



import lark
import pytest
from gotranx import atoms
from gotranx import exceptions
from gotranx import ode_component


def test_component_None(parser, trans):
    tree = parser.parse(
        (
            "parameters(x=1, y=2)\nparameters(z=3)\nstates(a=1)\n"
            "states(b=2, c=3)\nda_dt=0\ndb_dt=1\ndc_dt=2"
        ),
    )
    result = trans.transform(tree).components
    assert len(result) == 1  # Only one component
    comp = result[0]
    assert comp.name == ""
    assert comp.parameters == {
        atoms.Parameter(name="x", value=1),
        atoms.Parameter(name="y", value=2),
        atoms.Parameter(name="z", value=3),
    }
    assert comp.states == {
        atoms.State(name="a", value=1),
        atoms.State(name="b", value=2),
        atoms.State(name="c", value=3),
    }

    assert comp.assignments == {
        atoms.Assignment(
            name="da_dt",
            value=atoms.Expression(
                tree=lark.Tree("scientific", [lark.Token("SCIENTIFIC_NUMBER", "0")]),
            ),
        ),
        atoms.Assignment(
            name="db_dt",
            value=atoms.Expression(
                tree=lark.Tree("scientific", [lark.Token("SCIENTIFIC_NUMBER", "1")]),
            ),
        ),
        atoms.Assignment(
            name="dc_dt",
            value=atoms.Expression(
                tree=lark.Tree("scientific", [lark.Token("SCIENTIFIC_NUMBER", "2")]),
            ),
        ),
    }


@pytest.mark.parametrize(
    "index, name, num_parameters, num_states, num_assignments",
    [
        (0, "name", 2, 2, 2),
        (1, "a component", 0, 1, 1),
        (2, "b component", 0, 1, 1),
    ],
)
def test_component_with_multiple_component_names(
    index, name, num_parameters, num_states, num_assignments, parser, trans
):
    tree = parser.parse(
        'parameters("name", x=1, y=2)\n'
        'states("name", "a component", a=1)\n'
        'expressions("name", "a component")\n'
        "da_dt=0"
        'states("name", "b component", b=2)\n'
        'expressions("name", "b component")\n'
        "db_dt=0"
    )
    result = trans.transform(tree).components
    assert len(result) == 3

    assert result[index].name == name
    assert len(result[index].parameters) == num_parameters
    for p in result[index].parameters:
        assert name in p.components
    assert len(result[index].states) == num_states
    for s in result[index].states:
        assert name in s.components
    assert len(result[index].assignments) == num_assignments
    for a in result[index].assignments:
        assert name in a.components


def test_component_intermediates(parser, trans):
    tree = parser.parse(
        ("parameters(x=1, y=2)\n" "states(a=2, b=3)\n" "da_dt=0\n" "c=a+b\n" "db_dt=c - a"),
    )
    result = trans.transform(tree)
    comp = result.components[0]

    assert comp.intermediates == {
        atoms.Intermediate(
            name="c",
            value=atoms.Expression(
                tree=lark.Tree(
                    "add",
                    [
                        lark.Tree("variable", [lark.Token("VARIABLE", "a")]),
                        lark.Tree("variable", [lark.Token("VARIABLE", "b")]),
                    ],
                ),
            ),
            components=("",),
        ),
    }

    assert comp.state_derivatives == {
        atoms.StateDerivative(
            name="da_dt",
            value=atoms.Expression(
                tree=lark.Tree("scientific", [lark.Token("SCIENTIFIC_NUMBER", "0")]),
            ),
            components=("",),
            state=atoms.State(name="a", value=2.0, components=("",)),
        ),
        atoms.StateDerivative(
            name="db_dt",
            value=atoms.Expression(
                tree=lark.Tree(
                    "sub",
                    [
                        lark.Tree("variable", [lark.Token("VARIABLE", "c")]),
                        lark.Tree("variable", [lark.Token("VARIABLE", "a")]),
                    ],
                ),
            ),
            components=("",),
            state=atoms.State(name="b", value=3.0, components=("",)),
        ),
    }


@pytest.mark.parametrize(
    "expr, is_complete",
    [("states(a=2, b=3)\n da_dt=0\n db_dt=c - a", True), ("states(x=1)\n y=x", False)],
)
def test_component_is_complete(expr, is_complete, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    comp = result.components[0]
    assert comp.is_complete() is is_complete


def test_StateNotFound(parser, trans):
    expr = "states(x=1)\n y=x \n dy_dt=0"
    tree = parser.parse(expr)

    with pytest.raises(exceptions.StateNotFoundInComponent) as e:
        trans.transform(tree)

    assert str(e.value) == "State with name 'y' not found in component ''"


def test_state_with_and_without_derivatives(parser, trans):
    expr = "states(x=1, y=2)\n dy_dt=0"
    tree = parser.parse(expr)
    result = trans.transform(tree)
    comp = result.components[0]
    assert comp.states_with_derivatives == {
        atoms.State(name="y", value=2, components=("",)),
    }
    assert comp.states_without_derivatives == {
        atoms.State(name="x", value=1, components=("",)),
    }
    assert comp.is_complete() is False


@pytest.mark.parametrize(
    "expr, is_match, state_name",
    [
        ("dx_dt", True, "x"),
        ("dx_y_dt", True, "x_y"),
        ("dAb_x_t_dt", True, "Ab_x_t"),
        ("dx_dt_dt", True, "x_dt"),
        ("dxdt", False, ""),
        ("Dx_dt", False, ""),
        ("a_dx_dt", False, ""),
        ("adx_dt", False, ""),
        ("x_dt", False, ""),
        ("dx_dt_", False, ""),
        ("_dx_dt", False, ""),
        ("dx_dts", False, ""),
    ],
)
def test_STATE_DERIV_EXPR(expr, is_match, state_name):
    result = ode_component.STATE_DERIV_EXPR.match(expr)
    if is_match:
        assert result.groupdict()["state"] == state_name
    else:
        assert result is None



from __future__ import annotations

from collections import defaultdict
from typing import NamedTuple
from typing import Type
from typing import TypeVar

import lark

from . import atoms
from . import exceptions
from . import ode_component

T = TypeVar("T", atoms.State, atoms.Parameter)


class LarkODE(NamedTuple):
    components: tuple[ode_component.Component, ...]
    comments: tuple[atoms.Comment, ...]


def remove_quotes(s: str) -> str:
    return s.replace("'", "").replace('"', "")


def get_unit_from_assignment(s: lark.Tree) -> str | None:
    if len(s.children) >= 3:
        unit = s.children[2]
        try:
            if unit.type == "UNIT":
                return unit.value
        except AttributeError:
            pass
    return None


def find_assignments(
    s,
    components: tuple[str, ...],
) -> list[atoms.Assignment]:
    if isinstance(s, lark.Tree):
        return [
            atoms.Assignment(
                name=str(s.children[0]),
                value=atoms.Expression(tree=s.children[1]),
                components=tuple(components),
                unit_str=get_unit_from_assignment(s),
            ),
        ]

    return []


def find_components(
    s: list[None | lark.tree.Tree | lark.lexer.Token],
) -> tuple[int, tuple[str, ...]]:
    i = 0
    components = []

    while (
        isinstance(s[i], lark.lexer.Token)
        and hasattr(s[i], "type")
        and s[i].type == "COMPONENT_NAME"  # type:ignore
    ):
        components.append(remove_quotes(str(s[i])))
        i += 1

    # If no components are found, add an empty string
    if len(components) == 0:
        components.append("")
    return i, tuple(components)


def lark_list_to_parameters(
    s: list[None | lark.tree.Tree | lark.lexer.Token],
    cls: Type[T],
) -> tuple[T, ...]:
    i, components = find_components(s)
    return tuple(
        [
            tree2parameter(p, components=components, cls=cls)
            for p in s[i:]
            if isinstance(p, lark.Tree)
        ],
    )


def tree2parameter(
    s: lark.Tree,
    components: tuple[str, ...],
    cls: Type[T],
) -> T:
    from .expressions import build_expression

    if s.data == "param":
        return cls(
            name=str(s.children[0]),
            value=build_expression(s.children[1]),
            components=tuple(components),
        )
    elif s.data == "scalarparam":
        unit = None
        if s.children[2] is not None:
            unit = remove_quotes(str(s.children[2]))

        desc = None
        if s.children[3] is not None:
            desc = remove_quotes(str(s.children[3]))

        return cls(
            name=str(s.children[0]),
            value=build_expression(s.children[1]),
            components=tuple(components),
            unit_str=unit,
            description=desc,
        )
    else:
        raise exceptions.UnknownTreeTypeError(datatype=s.data, atom="Parameter")


class TreeToODE(lark.Transformer):
    def _call_userfunc(self, tree, new_children=None):  # pragma: nocover
        # Assumes tree is already transformed
        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            try:
                wrapper = getattr(f, "visit_wrapper", None)
                if wrapper is not None:
                    return f.visit_wrapper(f, tree.data, children, tree.meta)
                else:
                    return f(children)
            except lark.GrammarError:
                raise

    def _call_userfunc_token(self, token):  # pragma: nocover
        try:
            f = getattr(self, token.type)
        except AttributeError:
            return self.__default_token__(token)
        else:
            try:
                return f(token)
            except lark.GrammarError:
                raise

    def comment(self, s):
        return atoms.Comment(" ".join(map(str.lstrip, map(lambda x: x.lstrip("#"), map(str, s)))))

    def states(self, s: list[None | lark.tree.Tree | lark.lexer.Token]) -> tuple[atoms.State, ...]:
        return lark_list_to_parameters(s, cls=atoms.State)

    def parameters(
        self, s: list[None | lark.tree.Tree | lark.lexer.Token]
    ) -> tuple[atoms.Parameter, ...]:
        return lark_list_to_parameters(s, cls=atoms.Parameter)

    def expressions(self, s) -> tuple[atoms.Assignment, ...]:
        i, components = find_components(s)

        assignments = []

        for si in s[i:]:
            assignments.extend(find_assignments(si, components=components))

        return tuple(assignments)

    def ode(self, s) -> LarkODE:
        # FIXME: Could use Enum here
        mapping = {
            atoms.Parameter: "parameters",
            atoms.Assignment: "assignments",
            atoms.State: "states",
        }

        components: dict[str, dict[str, set[atoms.Atom]]] = defaultdict(
            lambda: {atom: set() for atom in mapping.values()},
        )

        # breakpoint()

        comments = []
        for line in s:  # Each line in the block
            if isinstance(line, atoms.Comment):
                comments.append(line)
                continue

            for atom in line:  # State, Parameters or Assignment
                for component in atom.components:
                    components[component][mapping[type(atom)]].add(atom)

        # Make sets frozen
        frozen_components: dict[str, dict[str, frozenset[atoms.Atom]]] = {}
        for component_name, component_values in components.items():
            frozen_components[component_name] = {}
            for atom_name, atom_values in component_values.items():
                frozen_components[component_name][atom_name] = frozenset(atom_values)

        # FIXME: Need to somehow tell the type checker that each of the inner dictionaries
        # are actually of the correct type.
        return LarkODE(
            components=tuple(
                [
                    ode_component.Component(name=key, **value)  # type: ignore
                    for key, value in frozen_components.items()
                ],
            ),
            comments=tuple(comments),
        )



from . import base
from . import c

from .c import CCodeGenerator
from .python import PythonCodeGenerator
from .base import CodeGenerator, Func, RHSArgument, SchemeArgument

__all__ = [
    "base",
    "c",
    "CCodeGenerator",
    "CodeGenerator",
    "Func",
    "RHSArgument",
    "SchemeArgument",
    "python",
    "PythonCodeGenerator",
]



from __future__ import annotations
from pathlib import Path
from typing import Any
from .cellml import CellMLParser


def cellml_to_gotran(filename: Path | str, params: dict[str, Any] | None = None) -> str:
    """Convert a cellml file to gotran code

    Parameters
    ----------
    input_filename : Path or str
        The path to the cellml file
    params : dict[str, Any], optional
        Parameters to pass to the parser, by default None

    Returns
    -------
    str
        The gotran code
    """
    parser = CellMLParser(filename, params=params)
    return parser.to_gotran()



from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import TypeVar, Callable
from functools import reduce

from structlog import get_logger

from . import atoms
from .ode import ODE
from .codegen.python import PythonCodeGenerator

T = TypeVar("T")
logger = get_logger()


def break_comment_at_80(acc, x):
    if x == "#":
        return acc + "\n#"

    if len(acc.split("\n")[-1]) + len(x) > 80:
        return acc + "\n# " + x
    else:
        return acc + " " + x


class GotranODECodePrinter(PythonCodeGenerator):
    def print_comments(self) -> str:
        if len(self.ode.comments) == 0:
            return ""
        text = "# "

        for comment in self.ode.comments:
            text += reduce(break_comment_at_80, comment.text.split(" "))

        text += "\n\n"
        return text

    def print_states(self) -> str:
        d: dict[tuple[str, ...], list[atoms.State]] = defaultdict(list)
        for state in self.ode.states:
            d[state.components].append(state)

        text = ""
        for components, states in d.items():
            text += start_odeblock("states", names=components) + "\n"
            text += ", ".join([print_ScalarParam(s, doprint=self.printer.doprint) for s in states])
            text += "\n)\n\n"
        return text

    def print_parameters(self) -> str:
        d: dict[tuple[str, ...], list[atoms.Parameter]] = defaultdict(list)
        for parameter in self.ode.parameters:
            d[parameter.components].append(parameter)

        text = ""
        for components, parameters in d.items():
            text += start_odeblock("parameters", names=components) + "\n"
            text += ",\n".join(
                [print_ScalarParam(p, doprint=self.printer.doprint) for p in parameters]
            )
            text += "\n)\n\n"

        return text

    def print_assignments(self) -> str:
        d: dict[tuple[str, ...], list[atoms.Assignment]] = defaultdict(list)
        for i in self.ode.intermediates + self.ode.state_derivatives:
            d[i.components].append(i)

        text = ""
        for components, intermediates in d.items():
            text += start_odeblock("expressions", names=components, is_expression=True) + "\n"
            text += "\n".join([print_assignment(i) for i in intermediates])
            text += "\n\n"

        return text

    def format(self, code: str) -> str:
        return self._formatter(code)


def start_odeblock(case: str, names: tuple[str, ...] = (), is_expression: bool = False):
    if len(names) == 0 or (len(names) == 1 and names[0] == ""):
        if is_expression:
            return ""
        return f"{case}("
    else:
        args = ", ".join(f'"{n}"' for n in names)

        if is_expression:
            return f"{case}({args})"
        else:
            return f"{case}({args},"


def join(*args: str) -> str:
    non_empty_args = [a for a in args if a != ""]
    if len(non_empty_args) == 0:
        return ""
    elif len(non_empty_args) == 1:
        return ", " + non_empty_args[0]
    else:
        return ", " + ", ".join(args)


def print_ScalarParam(p: atoms.Atom, doprint: Callable[[str], str]) -> str:
    unit_str = "" if p.unit_str is None else f'unit="{p.unit_str}"'
    description = "" if p.description is None else f'description="{p.description}"'
    if unit_str == "" and description == "":
        ret = f"{p.name}={doprint(p.value)}"  # type: ignore
    else:
        kwargs_str = join(unit_str, description)
        ret = f"{p.name}=ScalarParam({doprint(p.value)}{kwargs_str})"  # type: ignore

    logger.debug(f"Saving parameters {p} as {ret!r}")
    return ret


def print_assignment(a: atoms.Assignment) -> str:
    s = f"{a.name} = {a.expr}"
    if a.unit_str is not None:
        s += f" # {a.unit_str}"
    return s


def write_ODE_to_ode_file(ode: ODE, path: Path) -> None:
    # Just make sure it is a Path object
    printer = GotranODECodePrinter(ode)
    path = Path(path)
    text = ""
    text += printer.print_comments()
    text += printer.print_states()
    text += printer.print_parameters()
    text += printer.print_assignments()
    path.write_text(text)
    logger.info(f"Wrote {path}")



from __future__ import annotations

# Copyright (C) 2011-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.
import re
import urllib.error
import urllib.parse
import urllib.request
from collections import defaultdict
from collections import deque
from collections import OrderedDict
from functools import cmp_to_key
from pathlib import Path
from xml.etree import ElementTree
import structlog

logger = structlog.getLogger(__name__)

from .mathml import MathMLBaseParser


def cmp(a, b):
    return (a > b) - (a < b)


_all_keywords: list[str] = []


# Local imports

__all__ = ["CellMLParser"]

si_unit_map = {
    "ampere": "A",
    "becquerel": "Bq",
    "candela": "cd",
    "celsius": "gradC",
    "coulomb": "C",
    "dimensionless": "1",
    "farad": "F",
    "gram": "g",
    "gray": "Gy",
    "henry": "H",
    "hertz": "Hz",
    "joule": "J",
    "katal": "kat",
    "kelvin": "K",
    "kilogram": "kg",
    "liter": "l",
    "litre": "l",
    "molar": "M",
    "lumen": "lm",
    "lux": "lx",
    "meter": "m",
    "metre": "m",
    "mole": "mole",
    "newton": "N",
    "ohm": "Omega",
    "pascal": "Pa",
    "radian": "rad",
    "second": "s",
    "siemens": "S",
    "sievert": "Sv",
    "steradian": "sr",
    "tesla": "T",
    "volt": "V",
    "watt": "W",
    "weber": "Wb",
}

prefix_map = {
    "deca": "da",
    "hecto": "h",
    "kilo": "k",
    "mega": "M",
    "giga": "G",
    "tera": "T",
    "peta": "P",
    "exa": "E",
    "zetta": "Z",
    "yotta": "Y",
    "deci": "d",
    "centi": "c",
    "milli": "m",
    "micro": "u",
    "nano": "n",
    "pico": "p",
    "femto": "f",
    "atto": "a",
    "zepto": "z",
    "yocto": "y",
    None: "",
    "-3": "m",
}

ui = "UNINITIALIZED"


class Equation:
    """
    Class for holding information about an Equation
    """

    def __init__(self, name, expr, used_variables):
        self.name = name
        self.expr = expr
        self.used_variables = used_variables
        self.dependent_equations = []
        self.component = None

    def check_dependencies(self, equation):
        """
        Check Equation dependencies
        """
        assert isinstance(equation, Equation)

        if equation.name in self.used_variables:
            self.dependent_equations.append(equation)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Equation({self.name} = {''.join(self.expr)})"

    def __eq__(self, other):
        if not isinstance(other, Equation):
            return False
        return other.name == self.name and other.component == self.component

    def __hash__(self):
        return hash(self.name + self.component.name)


class Component:
    """
    Class for holding information about a CellML Component
    """

    def __init__(self, name, variables, equations, state_variables=None):
        self.name = name

        self.variable_info = {}

        self.state_variables = OrderedDict(
            (state, variables.pop(state, None)) for state in state_variables
        )

        for state, _info in list(self.state_variables.items()):
            self.variable_info[state] = _info
            self.variable_info["type"] = "state_variable"

        self.parameters = OrderedDict(
            (name, _info) for name, _info in list(variables.items()) if _info["init"] is not None
        )

        for param, _info in list(self.parameters.items()):
            self.variable_info[param] = _info
            self.variable_info["type"] = "parameter"

        self.derivatives = state_variables

        self.parent = None
        self.children = []

        self.dependencies = OrderedDict()
        self.used_in = OrderedDict()

        # Store equations
        self.sort_and_store_equations(equations)

        # Extract info
        dummy = dict(init=None, unit="1", private=True)
        for eq in self.equations:
            self.variable_info[eq.name] = variables.get(eq.name, dummy)
            self.variable_info[eq.name]["type"] = "equation"

        # Get used variables
        self.used_variables = set()
        for equation in self.equations:
            self.used_variables.update(equation.used_variables)

        # Remove dependencies on names defined by component
        self.used_variables.difference_update(equation.name for equation in self.equations)

        self.used_variables.difference_update(name for name in self.parameters)

        self.used_variables.difference_update(name for name in self.state_variables)

    def sort_and_store_equations(self, equations):
        # Check if reserved name for state derivativeis is used as equation
        # name
        derivative_names = [f"d{der}_dt" for der in self.derivatives]

        for eq in equations[:]:
            if eq.name in derivative_names and len(eq.expr) == 1 and eq.expr[0] == eq.name:
                equations.remove(eq)

        # Check internal dependencies
        for eq0 in equations:
            eq0.dependent_equations = []

            # Store component
            eq0.component = self
            for eq1 in equations:
                if eq0 == eq1:
                    continue
                eq0.check_dependencies(eq1)

        sorted_equations = []
        while equations:
            equation = equations.pop(0)
            if any(dep in equations for dep in equation.dependent_equations):
                equations.append(equation)
            else:
                sorted_equations.append(equation)

        # Store the sorted equations
        self.equations = sorted_equations

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name + f"<{len(self.state_variables)}>"

    def __repr__(self):
        return f"Component<{self.name}, {len(self.state_variables)}>"

    def __eq__(self, other):
        if not isinstance(other, Component):
            return False

        return other.name == self.name

    def check_dependencies(self, component):
        """
        Check components dependencies
        """
        assert isinstance(component, Component)

        if any(equation.name in self.used_variables for equation in component.equations):
            dep_equations = [
                equation for equation in component.equations if equation.name in self.used_variables
            ]

            # Register mutual dependencies
            self.dependencies[component] = dep_equations
            component.used_in[self] = dep_equations

            # Add logics for dependencies of all Equations.
            for other_equation in component.equations:
                for equation in self.equations:
                    if other_equation.name in equation.used_variables:
                        equation.dependent_equations.append(other_equation)

    def change_parameter_name(self, oldname, newname=None):
        """
        Change the name of a parameter
        Assume the name is only used locally within this component
        """
        assert oldname in self.parameters

        # If no newname is give we pick one based on the component name
        if newname is None:
            newname = oldname + "_" + self.name.split("_")[0]

        logger.warning(
            "Locally change parameter name '{0}' to '{1}' in " "component '{2}'.".format(
                oldname, newname, self.name
            ),
        )

        # Update parameters
        self.parameters = OrderedDict(
            (newname if name == oldname else name, value)
            for name, value in list(self.parameters.items())
        )

        # Update equations
        for eqn in self.equations:
            while oldname in eqn.expr:
                eqn.expr[eqn.expr.index(oldname)] = newname

        return newname

    def change_state_name(self, oldname, newname=None):
        """
        Change the name of a state
        Assume the name is only used locally within this component
        """
        assert oldname in self.state_variables

        # If no newname is give we pick one based on the component name
        if newname is None:
            newname = oldname + "_" + self.name.split("_")[0]

        logger.warning(
            "Locally change state name '{0}' to '{1}' in component " "'{2}'.".format(
                oldname, newname, self.name
            ),
        )

        # Update parameters
        self.state_variables = OrderedDict(
            (newname if name == oldname else name, value)
            for name, value in list(self.state_variables.items())
        )

        oldder = self.derivatives[oldname]
        newder = oldder.replace(oldname, newname)
        self.derivatives = OrderedDict(
            (newname if name == oldname else name, newder if value == oldder else value)
            for name, value in list(self.derivatives.items())
        )
        # Update equations
        for eqn in self.equations:
            while oldname in eqn.expr:
                eqn.expr[eqn.expr.index(oldname)] = newname
            while oldder in eqn.expr:
                eqn.expr[eqn.expr.index(oldder)] = newder
            if oldder == eqn.name:
                eqn.name = newder

        # Update derivative equation
        old_eq_name = f"d{oldname}_dt"
        new_eq_name = f"d{newname}_dt"
        self.variable_info[new_eq_name] = self.variable_info.pop(
            old_eq_name,
            dict(init=None, unit="1", private=True, type="equation"),
        )

        return newname

    def change_equation_name(self, oldname, newname=None):
        assert (
            oldname in self.variable_info and self.variable_info[oldname].get("type") == "equation"
        )

        # If no newname is give we pick one based on the component name
        if newname is None:
            newname = oldname + "_" + self.name.split("_")[0]

        logger.warning(
            "Locally change equation name '{0}' to '{1}' in " "component '{2}'.".format(
                oldname, newname, self.name
            ),
        )

        # Go through all equations and change the name used locally
        for ind, eqn in enumerate(self.equations[:]):
            while oldname in eqn.expr:
                eqn.expr[eqn.expr.index(oldname)] = newname

            # Update the actuall equation
            if eqn.name == oldname:
                eqn.name = newname
                self.equations[ind] = eqn

        # Update meta info
        self.variable_info[newname] = self.variable_info.pop(oldname)

        return newname

    def change_variable_name(self, oldname, newname=None):
        if oldname not in self.variable_info:
            logger.error(
                "Cannot change variable name. {0} is not a variable " "in component {1}".format(
                    oldname, self.name
                ),
            )
            return

        vartype = self.variable_info[oldname]["type"]

        assert vartype in ["state_variable", "parameter", "equation"]

        if vartype == "state_variable":
            return self.change_state_name(oldname, newname)
        if vartype == "parameter":
            return self.change_parameter_name(oldname, newname)

        return self.change_equation_name(oldname, newname)


class CellMLParser(object):
    """
    This module parses a CellML XML-file and converts it to PyCC code
    """

    @staticmethod
    def default_parameters():
        # Get the default parameters from the global parameters object
        return {
            "change_state_names": [],
            "grouping": "encapsulation",
        }

    def __init__(self, model_source, params=None, targets=None):
        """
        Arguments:
        ----------

        model_source: str
            Path or url to CellML file
        params : dict
            A dict with parameters for the
        targets : list, dict (optional)
            Components of the model to parse
        """

        targets = targets or []
        params = params or {}
        assert isinstance(model_source, (str, Path))
        assert isinstance(targets, (list, dict))
        self._params = self.default_parameters()
        self._params.update(params)

        # Open file or url
        try:
            with open(model_source, "r") as fp:
                self.cellml = ElementTree.parse(fp).getroot()
        except IOError:
            try:
                fp = urllib.request.urlopen(model_source)
                self.cellml = ElementTree.parse(fp).getroot()
            except Exception:
                logger.error("ERROR: Unable to open " + model_source)

        self.model_source = model_source
        self.name = self.cellml.attrib["name"]
        logger.info(f"Parsing CellML model: {self.name}")
        self.mathmlparser = MathMLBaseParser()
        self.cellml_namespace = self.cellml.tag.split("}")[0] + "}"
        self.parse_units()
        self.components = self.parse_components(targets)

        self.documentation = self.parse_documentation()

    def parse_documentation(self):
        """
        Parse the documentation of the article
        """
        namespace = "{http://cellml.org/tmp-documentation}"
        article = self.cellml.iter(namespace + "article")

        try:
            article = list(article)[0]
        except IndexError:
            return ""

        articleinfo = list(article.iter(namespace + "articleinfo"))[0]

        # Get title
        if articleinfo and articleinfo.iter(namespace + "title"):
            title = list(articleinfo.iter(namespace + "title"))[0].text
        else:
            title = ""

        # Get model structure comments
        for child in list(article):
            if child.attrib.get("id") == "sec_structure":
                content = []
                for par in child.iter(namespace + "para"):
                    # Get lines
                    splitted_line = deque(
                        ("".join(text.strip() for text in par.itertext())).split(" "),
                    )

                    # Cut them in lines which are not longer than 80 characters
                    ret_lines = []
                    while splitted_line:
                        line_stumps = []
                        line_length = 0
                        while splitted_line and (line_length + len(splitted_line[0]) < 80):
                            line_stumps.append(splitted_line.popleft())
                            line_length += len(line_stumps[-1]) + 1
                        ret_lines.append(" ".join(line_stumps))

                    content.extend(ret_lines)
                    content.append("\n")

                # Clean up content
                content = (
                    ("\n".join(cont.strip() for cont in content))
                    .replace("  ", " ")
                    .replace(" .", ".")
                    .replace(" ,", ",")
                )
                break
        else:
            content = ""

        if title or content:
            return f"{title}\n\n{content}"

        return ""

    def _gettag(self, node):
        """
        Splits off the namespace part from name, and returns the rest, the tag
        """
        return "".join(node.tag.split("}")[1:])

    def parse_units(self):
        """
        Parse any declared units in the model
        """
        collected_units = OrderedDict()
        unit_map = si_unit_map.copy()

        # Extend unit_map
        for cellml_pref, pref in list(prefix_map.items()):
            if cellml_pref:
                unit_map[cellml_pref + "molar"] = pref + "M"
                collected_units[cellml_pref + "molar"] = {pref + "M": (pref + "M", "1")}

        parsed_twice = []
        all_units = deque(self.get_iterator("units"))
        while all_units:
            units = all_units.popleft()
            unit_name = units.attrib["name"]

            if unit_name in unit_map:
                continue

            collected_parts = OrderedDict()
            for unit in list(units):
                if unit.attrib.get("multiplier"):
                    logger.warning(f"skipping multiplier in unit {units.attrib['name']}")
                if unit.attrib.get("multiplier"):
                    logger.warning(f"skipping multiplier in unit {units.attrib['name']}")
                cellml_unit = unit.attrib.get("units")

                prefix = prefix_map[unit.attrib.get("prefix")]
                exponent = unit.attrib.get("exponent", "1")
                if cellml_unit in si_unit_map:
                    abbrev = si_unit_map[cellml_unit]
                    name = prefix + abbrev
                    if exponent not in ["0", "1"]:
                        fullname = name + "**" + exponent
                    else:
                        fullname = name

                    collected_parts[name] = (fullname, exponent)
                elif cellml_unit in collected_units:
                    if prefix:
                        logger.warning(f"Skipping prefix of unit '{cellml_unit}'")
                    for name, (fullnam, part_exponent) in list(
                        collected_units[cellml_unit].items(),
                    ):
                        new_exponent = float(part_exponent) * float(exponent)

                        if new_exponent % 1.0 == 0.0:
                            new_exponent = str(int(new_exponent))
                        else:
                            new_exponent = str(new_exponent)

                        if new_exponent not in ["0", "1"]:
                            fullname = name + "**" + new_exponent
                        else:
                            fullname = name

                        collected_parts[name] = (fullname, exponent)

                elif units not in parsed_twice:
                    all_units.append(units)
                    parsed_twice.append(units)
                    break
                else:
                    logger.warning(
                        "Unknown unit '{0}' in {1}".format(cellml_unit, units["name"]),
                    )

            else:
                # Try change mole*l**-1 to mM...

                collected_units[unit_name] = collected_parts
                unit_map[unit_name] = "*".join(
                    fullname for fullname, exp in list(collected_parts.values())
                )

        # Store unit map
        self.units_map = unit_map

    def get_iterator(self, name, item=None):
        """
        Return an element tree iterator
        """

        item = item if item is not None else self.cellml
        return item.iter(self.cellml_namespace + name)

    def check_and_register_component_variables(
        self,
        comp,
        collected_states,
        collected_parameters,
        collected_equations,
    ):
        """
        Check if component variables are allready collected
        """

        # Check for duplication of states
        for name in list(comp.state_variables.keys()):
            der_name = f"d{name}_dt"
            if name in self._params["change_state_names"]:
                newname = name + "_" + comp.name.split("_")[0]
                comp.change_state_name(name, newname)
                name = newname

            # Check state name vs collected state names
            elif name in collected_states:
                state_comp = collected_states[name]
                logger.info(
                    "Same state name: '{0}' is used in component: '{1}' " "and '{2}'.".format(
                        name, comp.name, state_comp.name
                    ),
                )
                for change_comp in [comp, state_comp]:
                    if (
                        change_comp.state_variables[name]["private"]
                        and change_comp.variable_info[der_name]["private"]
                    ):
                        new_name = change_comp.change_state_name(name)
                        if change_comp == state_comp:
                            collected_states[new_name] = change_comp
                        break
                else:
                    logger.warning(
                        "Could not resolve duplicated state name {0} in "
                        "component {1} and {2}. None of them are private "
                        "to the components.".format(name, comp.name, state_comp.name),
                    )

            # Check state name vs collected parameter names
            elif name in collected_parameters:
                param_comp = collected_parameters[name]
                logger.info(
                    "State name: '{0}' from component '{1}' is used as "
                    "parameter in component '{2}'.".format(
                        name,
                        comp.name,
                        param_comp.name,
                    ),
                )

                # If parameter is private we change that
                if param_comp.parameters[name]["private"]:
                    new_name = param_comp.change_parameter_name(name)
                    collected_parameters.pop(name)
                    collected_parameters[new_name] = param_comp

                # Elseif the state variable is private we change that one
                elif (
                    comp.state_variables[name]["private"]
                    and comp.variable_info[der_name]["private"]
                ):
                    name = comp.change_state_name(name)

                else:
                    logger.warning(
                        "Could not resolve duplicated state and "
                        "parameter name {0} in component {1} and {2}. "
                        "None of them are private to the "
                        "components.".format(name, comp.name, param_comp.name),
                    )

            # Check state name vs collected equation names
            elif name in collected_equations:
                eq_comp = collected_equations[name]
                logger.info(
                    "State name '{0}' from component '{1}' is used as "
                    "parameter in component '{2}'.".format(
                        name,
                        comp.name,
                        eq_comp.name,
                    ),
                )

                # If state is private we change it
                if (
                    comp.state_variables[name]["private"]
                    and comp.variable_info[der_name]["private"]
                ):
                    name = comp.change_state_name(name)
                elif eq_comp.variable_info[name]["private"]:
                    new_name = eq_comp.change_equation_name(name)
                    collected_equations.pop(name)
                    collected_equations[new_name] = eq_comp

                else:
                    logger.warning(
                        "Could not resolve duplicated state and "
                        "equation name {0} in component {1} and {2}. "
                        "None of them are private to the "
                        "components.".format(name, comp.name, eq_comp.name),
                    )

            # Register the collected state
            collected_states[name] = comp

        # Check parameters
        for name in list(comp.parameters.keys()):
            # Check parameter name vs collected state names
            if name in collected_states:
                state_comp = collected_states[name]
                der_name = f"d{name}_dt"
                logger.info(
                    "Same parameter and state name: '{0}' is used in "
                    "component '{1}' and '{2}'.".format(
                        name,
                        comp.name,
                        state_comp.name,
                    ),
                )
                # If parameter is private we change that
                if comp.parameters[name]["private"]:
                    name = comp.change_parameter_name(name)
                elif (
                    state_comp.state_variables[name]["private"]
                    and state_comp.variable_info[der_name]["private"]
                ):
                    new_name = state_comp.change_state_name(name)
                    collected_states.pop(name)
                    collected_states[new_name] = state_comp

                else:
                    logger.warning(
                        "Could not resolve duplicated state name {0} in "
                        "component {1} and {2}. None of them are private "
                        "to the components.".format(name, comp.name, state_comp.name),
                    )

            # Check state name vs collected parameter names
            elif name in collected_parameters:
                param_comp = collected_parameters[name]
                logger.info(
                    "Parameter name '{0}' from component '{1}' is used as "
                    "parameter in component '{2}'.".format(
                        name,
                        comp.name,
                        param_comp.name,
                    ),
                )

                # If registered parameter is private we change that
                if param_comp.parameters[name]["private"]:
                    new_name = param_comp.change_parameter_name(name)
                    collected_parameters.pop(name)
                    collected_parameters[new_name] = param_comp

                # if the parameter is private we change that one
                elif comp.parameters[name]["private"]:
                    name = comp.change_parameter_name(name)

                else:
                    logger.warning(
                        "Could not resolve duplicated parameter names "
                        "{0} in component {1} and {2}. "
                        "None of them are private to the "
                        "components.".format(name, comp.name, param_comp.name),
                    )

            # Check state name vs collected equation names
            elif name in collected_equations:
                eq_comp = collected_equations[name]
                logger.info(
                    "Parameter name '{0}' from component '{1}' "
                    "is used as parameter in component '{2}'.".format(
                        name,
                        comp.name,
                        eq_comp.name,
                    ),
                )

                # If parameter is private we change it
                if comp.parameters[name]["private"]:
                    name = comp.change_parameter_name(name)
                elif eq_comp.variable_info[name]["private"]:
                    new_name = eq_comp.change_equation_name(name)
                    collected_equations.pop(name)
                    collected_equations[new_name] = eq_comp

                else:
                    logger.warning(
                        "Could not resolve duplicated parameter and "
                        "equation names {0} in component {1} and {2}. "
                        "None of them are private to the "
                        "components.".format(name, comp.name, eq_comp.name),
                    )
            collected_parameters[name] = comp

        # Check equation name
        for eq in comp.equations:
            name = eq.name

            # If the equation is private we do not care if it is duplicated
            # elsewhere. Risky...
            # if comp.variable_info[name]["private"]:
            #    collected_equations[name] = comp
            #    continue

            # If the equation will be broadcasted to another name we continue
            # print comp.name, name, self.new_variable_connections.get(comp.name, {}).keys()
            # if name in self.new_variable_connections.get(comp.name, {}).keys():
            #    continue

            # Check equation name vs collected state names
            if name in collected_states:
                state_comp = collected_states[name]
                der_name = f"d{name}_dt"
                logger.info(
                    "Same equation and state name '{0}' is used in "
                    "component '{1}' and '{2}'.".format(
                        name,
                        comp.name,
                        state_comp.name,
                    ),
                )
                # If equation is private we change that
                # if comp.variable_info[name]["private"]:
                #    name = comp.change_equation_name(name)
                if (
                    state_comp.state_variables[name]["private"]
                    and state_comp.variable_info[der_name]["private"]
                ):
                    new_name = state_comp.change_state_name(name)
                    collected_states.pop(name)
                    collected_states[new_name] = state_comp

                else:
                    logger.warning(
                        "Could not resolve duplicated state and "
                        "equation name {0} in component {1} and {2}. "
                        "None of them are private to the "
                        "components.".format(name, comp.name, state_comp.name),
                    )

            # Check equation name vs collected parameter names
            elif name in collected_parameters:
                param_comp = collected_parameters[name]
                logger.info(
                    "Equation name '{0}' from component '{1}' is used as "
                    "parameter in component '{2}'.".format(
                        name,
                        comp.name,
                        param_comp.name,
                    ),
                )

                # If registered parameter is private we change that
                if param_comp.parameters[name]["private"]:
                    new_name = param_comp.change_parameter_name(name)
                    collected_parameters.pop(name)
                    collected_parameters[new_name] = param_comp

                # If the parameter is private we change that one
                # elif comp.variable_info[name]["private"]:
                #    name = comp.change_equation_name(name)

                else:
                    logger.warning(
                        "Could not resolve duplicated parameter and "
                        "equation name {0} in component {1} and {2}. "
                        "None of them are private to the "
                        "components.".format(name, comp.name, param_comp.name),
                    )

            # Check equation name vs collected equation names
            elif name in collected_equations:
                eq_comp = collected_equations[name]
                logger.info(
                    "Equation name '{0}' from component '{1}' is used as "
                    "equation name in component '{2}'.".format(
                        name,
                        comp.name,
                        eq_comp.name,
                    ),
                )

                # If equation is private we change it
                # if comp.variable_info[name]["private"] and \
                #       eq_comp.variable_info[name]["private"]:
                #    warning("Both equations are private. We do not change "\
                #            "name on either one of them.")

                # elif eq_comp.variable_info[name]["private"]:
                #    new_name = eq_comp.change_equation_name(name)
                #    collected_equations.pop(name)
                #    collected_equations[new_name] = eq_comp
                #
                # else:
                #    warning("Could not resolve duplicated parameter and "\
                #            "equation names {0} in component {1} and {2}. "\
                #            "None of them are private to the "\
                #            "components.".format(name, comp.name, eq_comp.name))

            collected_equations[name] = comp

    def parse_imported_model(self):
        """
        Parse any imported models
        """

        components = OrderedDict()

        # Collect states and parameters to check for duplicates
        collected_states = dict()
        collected_parameters = dict()
        collected_equations = dict()

        # Import other models
        imports = self.get_iterator("import")
        if imports:
            logger.info("Parsing imported models:")
        for model in imports:
            import_comp_names = dict()

            for comp in self.get_iterator("component", model):
                import_comp_names[comp.attrib["component_ref"]] = comp.attrib["name"]

            model_parser = CellMLParser(
                model.attrib["{http://www.w3.org/1999/xlink}href"],
                targets=import_comp_names,
            )

            for comp in model_parser.components:
                components[comp.name] = comp

        # Extract states, parameters and equations
        for comp in list(components.values()):
            self.check_and_register_component_variables(
                comp,
                collected_states,
                collected_parameters,
                collected_equations,
            )

        return components, collected_states, collected_parameters, collected_equations

    def get_parents(self, grouping, element=None):
        """
        If group was used in the cellml use it to gather parent information
        about the components
        """

        # Collect component encapsulation
        def get_encapsulation(elements, all_parents, parent=None):
            children = {}
            for encap in elements:
                name = encap.attrib["component"]
                all_parents[name] = parent
                if list(encap):
                    nested_children = get_encapsulation(list(encap), all_parents, name)
                    children[name] = dict(children=nested_children, parent=parent)
                else:
                    children[name] = dict(children=None, parent=parent)

            return children

        encapsulations = dict()
        all_parents = dict()
        for group in self.get_iterator("group", element):
            children = list(group)

            if children and children[0].attrib.get("relationship") == grouping:
                encapsulations = get_encapsulation(children[1:], all_parents)

        # If no group information in cellml extract potential parent information
        # from component names
        if not all_parents:
            # Iterate over the components
            comp_names = [comp.attrib["name"] for comp in self.get_iterator("component", element)]

            for parent_name in comp_names:
                for name in comp_names:
                    if parent_name in name and parent_name != name:
                        all_parents[name] = parent_name

        return encapsulations, all_parents

    def parse_single_component(
        self,
        comp,
        collected_states,
        collected_parameters,
        collected_equations,
    ):
        """
        Parse a single component and create a Component object
        """
        comp_name = comp.attrib["name"]

        # Collect variables and equations
        variables = OrderedDict()
        equations = []
        state_variables = OrderedDict()
        # derivatives = []

        # Get variables that are used outside the component
        variables_used_in_connections = list(
            self.new_variable_connections.get(comp_name, dict()).keys(),
        ) + list(self.same_variable_connections.get(comp_name, dict()).keys())

        # Get variable and initial values
        for var in self.get_iterator("variable", comp):
            var_name = var.attrib["name"]
            if var_name in _all_keywords:
                var_name = var_name + "_"

            # Check connection and pair it with interface information to
            # check if a variable is public or not
            public = var_name in variables_used_in_connections and (
                var.attrib.get("public_interface") == "out"
                or var.attrib.get("private_interface") == "out"
            )

            # Store variables using initial and unit
            variables[var_name] = dict(
                init=var.attrib.get("initial_value"),
                unit=self.units_map[var.attrib.get("units")],
                private=not public,
                public_interface=var.attrib.get("public_interface"),
                private_interface=var.attrib.get("private_interface"),
            )

        # Get equations
        for math in comp.iter("{http://www.w3.org/1998/Math/MathML}math"):
            for eq in list(math):
                (
                    equation_list,
                    state_variable,
                    derivative,
                    used_variables,
                ) = self.mathmlparser.parse(eq)

                # Get equation name
                eq_name = equation_list[0]

                if eq_name in _all_keywords:
                    equation_list[0] = eq_name + "_"
                    eq_name = equation_list[0]

                # Discard collected equation name from used variables
                used_variables.discard(eq_name)

                assert re.findall(r"(\w+)", eq_name)[0] == eq_name
                assert equation_list[1] == self.mathmlparser["eq"]
                equations.append(Equation(eq_name, equation_list[2:], used_variables))

                # Do not register state variables twice
                if state_variable is not None and state_variable not in state_variables:
                    state_variables[state_variable] = derivative

        # Create Component
        comp = Component(comp_name, variables, equations, state_variables)

        # Collect and check variables
        self.check_and_register_component_variables(
            comp,
            collected_states,
            collected_parameters,
            collected_equations,
        )

        return comp

    def sort_components(self, components, sorted_once=False):
        if not sorted_once:
            logger.info("Sorting components with respect to dependencies")
        else:
            logger.info("Sorting components with respect to dependencies second time")

        # Check internal dependencies
        for comp0 in components:
            for comp1 in components:
                if comp0 == comp1:
                    continue
                comp0.check_dependencies(comp1)

        def simple_sort(components):
            components = deque(components)
            dependant_components = []
            sorted_components = []
            while components:
                component = components.popleft()

                # Chek for circular dependancy
                if dependant_components.count(component) > 4:
                    components.append(component)
                    break

                if any(dep in components for dep in component.dependencies):
                    components.append(component)
                    dependant_components.append(component)
                else:
                    sorted_components.append(component)

            return sorted_components, list(components)

        # Initial sorting
        sorted_components, circular_components = simple_sort(components)

        # If no circular dependencies
        if not circular_components:
            return sorted_components

        try:
            import networkx as nx
        except ImportError:
            logger.warning(
                "networkx could not be imported. Circular "
                "dependencies between components will not be sorted out.",
            )
            return sorted_components + circular_components

        # Collect zero and one dependencies
        zero_dep_equations = set()
        one_dep_equations = set()
        equation_map = {}

        # Gather zero dependent equations
        for comp in circular_components:
            for dep_comp, equations in list(comp.dependencies.items()):
                for equation in equations:
                    if not equation.dependent_equations and equation.name in comp.used_variables:
                        zero_dep_equations.add(equation)
                        equation_map[equation.name] = equation

        # Check for one dependency if that is the zero one
        one_dep_zero_dep = defaultdict(set)
        for comp in circular_components:
            for dep_comp, equations in list(comp.dependencies.items()):
                for equation in equations:
                    if (
                        len(equation.dependent_equations) == 1
                        and equation.name in comp.used_variables
                        and equation.dependent_equations[0] in zero_dep_equations
                    ):
                        equation_map[equation.name] = equation
                        one_dep_equations.add(equation)
                        one_dep_zero_dep[equation.name].add(
                            equation.dependent_equations[0].name,
                        )

        # Try to eliminate circular dependency
        # Extract dependent equation to a new component

        # Find ODE component. If not found create one
        for comp in components:
            if comp.name == self.name:
                ode_comp = comp
                break
        else:
            ode_comp = Component(self.name, {}, [], {})

        # Valid edges for removal
        valid_edges = [eq.name for eq in zero_dep_equations] + [eq.name for eq in one_dep_equations]

        G = nx.MultiDiGraph()
        G.add_nodes_from([comp.name for comp in components])

        # Build graph
        for comp in components:
            [
                G.add_edge(dep.name, comp.name, key=equation.name)
                for dep, equations in list(comp.dependencies.items())
                for equation in equations
            ]

        # collecte edges that breaks cycles
        cycle_breakers = []

        # Collect data over the best edge to remove
        edge_score = defaultdict(lambda: 0)
        edge_to_nodes = defaultdict(set)
        for cycle in nx.simple_cycles(G):
            local_breaker = []
            for n0, n1 in zip(cycle[:-1], cycle[1:]):
                if all(edge in valid_edges for edge in G[n0][n1]):
                    local_breaker.append([edge for edge in G[n0][n1]])
                    for edge in local_breaker[-1]:
                        edge_to_nodes[edge].add((n0, n1))
            cycle_breakers.append(local_breaker)
            for local_edges in local_breaker:
                for edge in local_edges:
                    edge_score[edge] += 1

        # Sort the edges we should remove by a score given by how many
        # cycles it breaks by being removed.
        edge_score = sorted((edge_score[edge], edge) for edge in edge_score)

        # Collect edges to be removed
        edge_removal = set()
        cycles_fixed = [False] * len(cycle_breakers)

        while not all(cycles_fixed) and edge_score:
            score, edge = edge_score.pop()

            # Remove this/these edge/edges this iteration
            local_removal = [edge]

            # If adding a one dep edge we need to also remove its dependent edge
            dep_removal = set()
            if edge in one_dep_zero_dep:
                local_removal.extend(one_dep_zero_dep[edge])
                dep_removal.update(one_dep_zero_dep[edge])

            for edge_remove in local_removal:
                # Iterate over the different cycles
                for i, local_breakers in enumerate(cycle_breakers):
                    # If the cycle is fixed
                    if cycles_fixed[i]:
                        continue

                    # Go through the collected valid breakers and pick the one
                    # with least edges first
                    for j, local_edges in enumerate(
                        sorted(
                            local_breakers,
                            key=cmp_to_key(lambda o0, o1: cmp(len(o0), len(o1))),
                        ),
                    ):
                        # If the removed edge is in the local edges
                        if edge_remove in local_edges:
                            local_edges.remove(edge_remove)
                            edge_removal.add(edge_remove)

                            # Check any dependent edges
                            for dep_edge in one_dep_zero_dep.get(edge_remove, []):
                                edge_removal.add(dep_edge)

                            if len(local_edges) == 0:
                                cycles_fixed[i] = True

                                break

        # If no edge sugested for removal we pick the zero dep equations
        if not edge_removal:
            if zero_dep_equations:
                edge_removal = [eq.name for eq in zero_dep_equations]
            elif one_dep_equations:
                edge_removal = [eq.name for eq in one_dep_equations]
            else:
                logger.warning("Could not sort out circular dependencies.")

        # Remove the edges from the graph
        for edge in edge_removal:
            for n0, n1 in edge_to_nodes[edge]:
                G.remove_edge(n0, n1, key=edge)

        # Move the marked edges (equations) from the relevant components
        removed_equations = {}
        for edge in edge_removal:
            eq = equation_map[edge]

            old_comp = eq.component
            ode_comp.equations.append(eq)
            old_comp.equations.remove(eq)
            ode_comp.variable_info[eq.name] = old_comp.variable_info.pop(eq.name)

            # Store changed
            removed_equations[eq] = old_comp

            # Transfer used_in to new component
            new_dependent_componets = OrderedDict()
            for dep_comp, equations in list(old_comp.used_in.items()):
                assert equations

                if eq not in equations or dep_comp == ode_comp:
                    continue

                # Remove dependency from old component and add it to the new
                if len(equations) == 1:
                    new_dependent_componets[dep_comp] = old_comp.used_in.pop(dep_comp)
                else:
                    new_dependent_componets[dep_comp] = [
                        old_comp.used_in[dep_comp].pop(equations.index(eq)),
                    ]

                # Change component dependencies
                if old_comp in dep_comp.dependencies and eq in dep_comp.dependencies[old_comp]:
                    if len(dep_comp.dependencies[old_comp]) == 1:
                        dep_comp.dependencies.pop(old_comp)
                    else:
                        dep_comp.dependencies[old_comp].remove(eq)

                if ode_comp not in dep_comp.dependencies:
                    dep_comp.dependencies[ode_comp] = [eq]
                else:
                    dep_comp.dependencies[ode_comp].append(eq)

            # Store new component to the extracted equation
            eq.component = ode_comp
            ode_comp.dependent_componets = new_dependent_componets

            if ode_comp not in sorted_components:
                sorted_components.insert(0, ode_comp)

        # Sort newly added equations
        ode_comp.sort_and_store_equations(ode_comp.equations)

        # Sort graph components and apply the sortings to the collected
        # CellML compoents
        sort_again = False
        try:
            sorted_components = list(nx.topological_sort(G))
            components.sort(key=lambda n0: sorted_components.index(n0.name))
            message = "To avoid circular dependency the following equations " "has been moved:"

        except nx.NetworkXUnfeasible as e:
            logger.warning("Topological sort failed: " + str(e))
            message = (
                "In a try to avoid circular dependency the following equations " "has been moved:"
            )
            sort_again = True

        # Insert the ODE component with extracted equations first
        components.insert(0, ode_comp)

        logger.warning(message)

        for eq, old_comp in list(removed_equations.items()):
            logger.warning(f"{eq.name} : from {old_comp.name} to {ode_comp.name} component")

        if sort_again:
            # Try rebuild the graph and make another topological sort
            G = nx.MultiDiGraph()
            G.add_nodes_from([comp.name for comp in components])

            # Build graph
            for comp in components:
                [
                    G.add_edge(dep.name, comp.name, key=equation.name)
                    for dep, equations in list(comp.dependencies.items())
                    for equation in equations
                ]

            try:
                sorted_components = list(nx.topological_sort(G))
                components.sort(key=lambda n0: sorted_components.index(n0.name))
            except nx.NetworkXUnfeasible as e:
                logger.warning("Topological sort failed a second time: " + str(e))

        return components

    def parse_components(self, targets):
        """
        Build a dictionary containing dictionarys describing each
        component of the cellml model
        """

        # First parse connections which is used to determine the interface of
        # each variable
        (
            self.new_variable_connections,
            self.same_variable_connections,
        ) = self.parse_connections()

        # Parse imported components
        (
            components,
            collected_states,
            collected_parameters,
            collected_equations,
        ) = self.parse_imported_model()

        # Get parent relationship between components
        encapsulations, all_parents = self.get_parents(self._params["grouping"])

        if targets:
            # If the parent information was not of type encapsulation
            # regather parent information
            if self._params["grouping"] != "encapsulation":
                encapsulations, dummy = self.get_parents("encapsulation")

            # Add any encapsulated components to the target list
            for target, new_target_name in list(targets.items()):
                if target in encapsulations:
                    for child in encapsulations[target]["children"]:
                        targets[child] = child.replace(target, new_target_name)

            target_parents = dict()

            # Update all_parents
            for comp_name, parent_name in list(all_parents.items()):
                if parent_name not in targets:
                    continue
                target_parents[targets[comp_name]] = targets[parent_name]

        # Iterate over the components
        for comp in self.get_iterator("component"):
            comp_name = comp.attrib["name"]

            # Only parse selected and non-empty components
            if (targets and comp_name not in targets) or len(list(comp)) == 0:
                continue

            # If targets provides a name mapping give the component a new name
            if targets and isinstance(targets, dict):
                new_name = targets[comp_name]
                comp.attrib["name"] = new_name
                comp_name = new_name

            # Store component
            components[comp_name] = self.parse_single_component(
                comp,
                collected_states,
                collected_parameters,
                collected_equations,
            )

        # Add parent information
        all_component_names = list(components.keys())
        for name, comp in list(components.items()):
            if targets:
                parent_name = target_parents.get(name)
            else:
                parent_name = all_parents.get(name)

            if parent_name:
                comp.parent = components[parent_name]

                # If parent name in child name, reduce child name length
                if parent_name in comp.name:
                    old_name = comp.name
                    new_name = old_name.replace(parent_name, "").strip("_")
                    if new_name not in all_component_names:
                        comp.name = new_name
                        all_component_names.remove(old_name)
                        all_component_names.append(new_name)

                components[parent_name].children.append(comp)

        # If we only extract a sub set of component we do not sort
        if targets:
            return list(components.values())

        # Before dependencies are checked we change names according to
        # variable mappings in the original CellML file
        if self.new_variable_connections:
            logger.info("\nRenaming variables through connections")

        for comp, variables in list(self.new_variable_connections.items()):
            # Iterate over old and new names
            for oldname, newnames in list(variables.items()):
                # Check that the direction of the connection is correct
                # If no variable info is registered for this comp or its
                # children it is the wrong direction

                # print
                # print comp, oldname
                # print components[comp].variable_info.keys()
                # for child in components[comp].children:
                #    print child.name, child.variable_info.keys()
                #    if oldname in child.variable_info.keys():
                #        print "HURRAY"
                #        print child.variable_info[oldname]

                if oldname not in components[comp].variable_info and all(
                    oldname not in child.variable_info for child in components[comp].children
                ):
                    continue

                # Try access var_info
                var_info = components[comp].variable_info.get(oldname)

                # If not we try components children
                if var_info is None:
                    for child in components[comp].children:
                        if oldname in child.variable_info:
                            var_info = child.variable_info[oldname]
                            break

                # If variable is not intended out
                if not (
                    var_info.get("public_interface") == "out"
                    or var_info.get("private_interface") == "out"
                ):
                    continue

                # Check if the oldname is used in any components
                old_name_used = self.same_variable_connections[comp].get(oldname)

                # If there are only one newname we change the name of the
                # original equation
                if len(newnames) == 1:
                    # If not old name used or old name only used in children and
                    # we then assume it is private to the component relationship
                    # parent-child
                    if not old_name_used or all(
                        components[other_comp] in components[comp].children
                        for other_comp in old_name_used
                    ):
                        # Get new name
                        newname = list(newnames.keys())[0]

                        # If oldname in component
                        if components[comp].variable_info.get(oldname) is not None:
                            components[comp].change_variable_name(oldname, newname)

                        if old_name_used:
                            for other_comp in old_name_used:
                                components[other_comp].change_variable_name(
                                    oldname,
                                    newname,
                                )

                        for child in components[comp].children:
                            if child.variable_info.get(oldname) is not None:
                                child.change_variable_name(oldname, newname)
                    else:
                        # FIXME: Add equation with name change to component
                        pass
                else:
                    # FIXME: Add equation with name change to component
                    pass

        # Add dependencies and sort the components accordingly
        return self.sort_components(list(components.values()))

    def parse_connections(self):
        new_variable_names = dict()
        same_variable_names = dict()

        for con in self.get_iterator("connection"):
            con_map = list(self.get_iterator("map_components", con))[0]
            comp1 = con_map.attrib["component_1"]
            comp2 = con_map.attrib["component_2"]

            for var_map in self.get_iterator("map_variables", con):
                var1 = var_map.attrib["variable_1"]
                var2 = var_map.attrib["variable_2"]

                if var1 != var2:
                    if comp1 not in new_variable_names:
                        new_variable_names[comp1] = {var1: defaultdict(list)}
                    elif var1 not in new_variable_names[comp1]:
                        new_variable_names[comp1][var1] = defaultdict(list)
                    new_variable_names[comp1][var1][var2].append(comp2)

                    # Register both directions
                    if comp2 not in new_variable_names:
                        new_variable_names[comp2] = {var2: defaultdict(list)}
                    elif var2 not in new_variable_names[comp2]:
                        new_variable_names[comp2][var2] = defaultdict(list)
                    new_variable_names[comp2][var2][var1].append(comp1)

                else:
                    if comp1 not in same_variable_names:
                        same_variable_names[comp1] = {var1: [comp2]}
                    elif var1 not in same_variable_names[comp1]:
                        same_variable_names[comp1][var1] = [comp2]
                    else:
                        same_variable_names[comp1][var1].append(comp2)

                    # Register both directions
                    if comp2 not in same_variable_names:
                        same_variable_names[comp2] = {var2: [comp1]}
                    elif var2 not in same_variable_names[comp2]:
                        same_variable_names[comp2][var2] = [comp1]
                    else:
                        same_variable_names[comp2][var2].append(comp1)

        return new_variable_names, same_variable_names

    def to_gotran(self):
        """
        Generate a gotran file
        """
        gotran_lines = []
        for docline in self.documentation.split("\n"):
            gotran_lines.append("# " + docline)

        if gotran_lines:
            gotran_lines.extend([""])

        # Add component info
        declaration_lines = []
        equation_lines = []

        def unders_score_replace(comp):
            new_name = comp.name.replace("_", " ")

            # If only 1 state it might be included in the name
            state_name = ""
            if len(comp.state_variables) == 1:
                state_name = list(comp.state_variables.keys())[0]
                if state_name in comp.name:
                    new_name = state_name.join(
                        part.replace("_", " ") for part in comp.name.split(state_name)
                    )

            # Captitalize first word
            single_words = new_name.split(" ")
            if len(single_words[0]) > 1 and "_" not in single_words[0]:
                single_words[0] = single_words[0][0].upper() + single_words[0][1:]

            # If first word is only a character we assume the first two
            # words to stick together
            elif (
                len(single_words[0]) == 1
                and len(single_words) > 1
                and single_words[0] != state_name
            ):
                single_words = [single_words[0] + "_" + single_words[1]] + single_words[2:]

            return " ".join(single_words)

        # Iterate over components and collect stuff
        for comp in self.components:
            names = deque([unders_score_replace(comp)])

            parent = comp.parent
            while parent is not None:
                names.appendleft(unders_score_replace(parent))
                parent = parent.parent

            comp_name = ", ".join(f'"{name}"' for name in names)

            # Collect initial state values
            if comp.state_variables:
                declaration_lines.append("")
                declaration_lines.append(f"states({comp_name},")
                for name, _info in list(comp.state_variables.items()):
                    if _info["unit"] != "1":
                        declaration_lines.append(
                            "       {0} = ScalarParam({1}" ', unit="{2}"),'.format(
                                name,
                                float(_info["init"]),
                                _info["unit"],
                            ),
                        )
                    else:
                        declaration_lines.append(f"       {name} = {_info['init']},")
                declaration_lines[-1] = declaration_lines[-1][:-1] + ")"

            # Collect initial parameters values
            if comp.parameters:
                declaration_lines.append("")
                declaration_lines.append(f"parameters({comp_name},")
                for name, _info in list(comp.parameters.items()):
                    if _info["unit"] != "1":
                        declaration_lines.append(
                            "           {0} = ScalarParam({1}" ', unit="{2}"),'.format(
                                name,
                                float(_info["init"]),
                                _info["unit"],
                            ),
                        )
                    else:
                        declaration_lines.append(
                            f"           {name} = {_info['init']},",
                        )
                declaration_lines[-1] = declaration_lines[-1][:-1] + ")"

            # Collect all intermediate equations
            if comp.equations:
                equation_lines.append("")
                equation_lines.append(f"expressions({comp_name})")
                """
                for eq in comp.equations:
                    expr = []
                    for eqi in eq.expr:
                        if eqi.isdigit():
                            # This converts integers to floats
                            expr.append(str(float(eqi)))
                        else:
                            expr.append(eqi)
                    eq.expr = expr
                """

                equation_lines.extend(
                    "{0} = {1}{2}".format(
                        eq.name,
                        "".join(eq.expr),
                        " # {0}".format(comp.variable_info[eq.name]["unit"])
                        if comp.variable_info[eq.name]["unit"] != "1"
                        else "",
                    )
                    for eq in comp.equations
                )

        gotran_lines.append(
            f"# gotran file generated by cellml2gotran from {self.model_source}",
        )
        gotran_lines.extend(declaration_lines)
        gotran_lines.extend(equation_lines)
        gotran_lines.append("")
        gotran_lines.append("")

        # Return joined lines
        return "\n".join(gotran_lines)



import gotranx
import pytest
from pathlib import Path
from typer.testing import CliRunner

here = Path(__file__).parent.absolute()
runner = CliRunner(mix_stderr=False)


@pytest.fixture(scope="module")
def odefile(tmp_path_factory):
    expr = """
    parameters(a=0)
    parameters("My component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states("My component", x=1.0, y=2.0,z=3.05)

    expressions("My component")
    dy_dt = x*(rho - z) - y # millivolt
    dx_dt = sigma*(-x + y)
    dz_dt = -beta*z + x*y
    """
    fname = tmp_path_factory.mktemp("data") / "lorentz.ode"
    fname.write_text(expr)
    yield fname
    fname.unlink()


def test_cli_version():
    result = runner.invoke(gotranx.cli.app, "--version")
    assert result.exit_code == 0
    assert "gotranx" in result.stdout


def test_cli_license():
    result = runner.invoke(gotranx.cli.app, "--license")
    assert result.exit_code == 0
    assert "MIT" in result.stdout


def test_gotran2py(odefile):
    outfile = odefile.with_suffix(".py")
    result = runner.invoke(gotranx.cli.app, [str(odefile), "--to", ".py", "-o", str(outfile)])
    assert result.exit_code == 0
    assert "lorentz.py" in result.stdout
    assert "lorentz" in result.stdout
    assert outfile.is_file()
    outfile.unlink()


def test_gotran2c(odefile):
    outfile = odefile.with_suffix(".h")
    result = runner.invoke(gotranx.cli.app, [str(odefile), "--to", ".h", "-o", str(outfile)])
    assert result.exit_code == 0
    assert "lorentz.h" in result.stdout
    assert "lorentz" in result.stdout
    assert outfile.is_file()
    outfile.with_suffix(".h").unlink()


def test_cellml2ode():
    cellmlfile = here / "cellml_files" / "noble_1962.cellml"
    out_odefile = cellmlfile.with_suffix(".ode")
    result = runner.invoke(gotranx.cli.app, [str(cellmlfile), "-o", cellmlfile.with_suffix(".ode")])
    assert result.exit_code == 0

    assert f"Wrote {out_odefile}" in result.stdout
    assert out_odefile.is_file()
    # Check that we can load the file
    ode = gotranx.load_ode(out_odefile)
    assert ode is not None
    out_odefile.unlink()



import pytest
from gotranx import atoms


@pytest.mark.parametrize(
    "expr",
    [
        "states(x=1)",
        "states(x = 1)",
        "states(x= 1)",
        """states(
    x=1)""",
        """states(
    x=1
    )""",
    ],
)
def test_states_single(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 1
    assert result[0] == atoms.State(name="x", value=1)


@pytest.mark.parametrize(
    "expr",
    [
        "states(x=1, y=2)",
        "states(x = 1, y =2)",
        "states(x= 1,y= 2)",
        """states(
    x=1, y=2)""",
        """states(
    x=1,
    y=2
    )""",
    ],
)
def test_states_double(expr, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert len(result) == 2

    assert result[0] == atoms.State(name="x", value=1)
    assert result[1] == atoms.State(name="y", value=2)


def test_states_with_component(parser, trans):
    expr = 'states("My component", x=1, y=2)'
    tree = parser.parse(expr)
    result = trans.transform(tree)

    assert len(result) == 2

    assert result[0] == atoms.State(name="x", value=1, components=("My component",))
    assert result[1] == atoms.State(name="y", value=2, components=("My component",))


def test_states_with_component_and_info(parser, trans):
    expr = 'states("My component", "Some info about the component", x=1, y=2)'
    tree = parser.parse(expr)
    result = trans.transform(tree)

    assert len(result) == 2

    assert result[0] == atoms.State(
        name="x",
        value=1,
        components=(
            "My component",
            "Some info about the component",
        ),
    )
    assert result[1] == atoms.State(
        name="y",
        value=2,
        components=(
            "My component",
            "Some info about the component",
        ),
    )


def test_different_sets_of_states(parser, trans):
    expr = (
        'states("First component", "Some info about first '
        'component", x=1, y=2)\nstates("Second component", z=3)\nstates(w=4)'
    )
    tree = parser.parse(expr)
    result = trans.transform(tree).components

    assert len(result) == 4
    first_component = result[0]

    assert first_component.states == result[1].states
    assert first_component.name == "First component"
    assert first_component.states == {
        atoms.State(
            name="x",
            value=1,
            components=(
                "First component",
                "Some info about first component",
            ),
        ),
        atoms.State(
            name="y",
            value=2,
            components=(
                "First component",
                "Some info about first component",
            ),
        ),
    }

    second_component = result[2]
    assert second_component.name == "Second component"
    assert second_component.states == {
        atoms.State(name="z", value=3, components=("Second component",)),
    }

    third_component = result[3]
    assert third_component.name == ""
    assert third_component.states == {
        atoms.State(name="w", value=4, components=("",)),
    }


@pytest.mark.parametrize(
    "expr, expected",
    [
        (
            'states(x=ScalarParam(1, unit="pA"))',
            (atoms.State(name="x", value=1.0, unit_str="pA"),),
        ),
        (
            'states(x=ScalarParam(1, unit="pA", description="Info about x"))',
            (
                atoms.State(
                    name="x",
                    value=1.0,
                    unit_str="pA",
                    description="Info about x",
                ),
            ),
        ),
        (
            'states(x=ScalarParam(1, unit="pA*pF**-1", description="Info about x"))',
            (
                atoms.State(
                    name="x",
                    value=1.0,
                    unit_str="pA*pF**-1",
                    description="Info about x",
                ),
            ),
        ),
        (
            """
        states(
            x=ScalarParam(1, unit="pA", description="Info about x"),
            y=ScalarParam(2.0, unit="mM", description="Info about y")
        )
        """,
            (
                atoms.State(
                    name="x",
                    value=1.0,
                    unit_str="pA",
                    description="Info about x",
                ),
                atoms.State(
                    name="y",
                    value=2.0,
                    unit_str="mM",
                    description="Info about y",
                ),
            ),
        ),
    ],
)
def test_states_with_unit_and_desc(expr, expected, parser, trans):
    tree = parser.parse(expr)
    result = trans.transform(tree)
    assert result == expected



from __future__ import annotations
from pathlib import Path
from structlog import get_logger

from ..codegen.c import CCodeGenerator
from ..load import load_ode
from ..schemes import Scheme

logger = get_logger()


def main(
    fname: Path,
    suffix: str = ".h",
    outname: str | None = None,
    scheme: list[Scheme] | None = None,
) -> None:
    ode = load_ode(fname)
    codegen = CCodeGenerator(ode)
    comp = [
        "#include <math.h>",
        codegen.initial_parameter_values(),
        codegen.initial_state_values(),
        codegen.rhs(),
    ]
    if scheme is not None:
        for s in scheme:
            comp.append(codegen.scheme(s.value))

    code = "\n".join(comp)
    out = fname if outname is None else Path(outname)
    out_name = out.with_suffix(suffix=suffix)
    out_name.write_text(codegen._format(code))
    logger.info(f"Wrote {out_name}")



from textwrap import dedent, indent


def init_state_values(name, state_names, state_values, code):
    indented_code = indent(code, "    ")
    indent_values = indent(", ".join(f"{n}={v}" for n, v in zip(state_names, state_values)), "    ")
    return dedent(
        f"""
void init_state_values(double* {name}){{
    /*
{indent_values}
    */
{indented_code}
}}
""",
    )


def init_parameter_values(name, parameter_names, parameter_values, code):
    indented_code = indent(code, "    ")
    indent_values = indent(
        ", ".join(f"{n}={v}" for n, v in zip(parameter_names, parameter_values)), "    "
    )
    return dedent(
        f"""
void init_parameter_values(double* {name}){{
    /*
{indent_values}
    */
{indented_code}
}}
""",
    )


def method(
    name: str,
    args: str,
    states: str,
    parameters: str,
    values: str,
    return_name: None = None,
    num_return_values: int = 0,
):
    indent_states = indent(states, "    ")
    indent_parameters = indent(parameters, "    ")
    indent_values = indent(values, "    ")
    return dedent(
        f"""
void {name}({args}){{

    // Assign states
{indent_states}

    // Assign parameters
{indent_parameters}

    // Assign expressions
{indent_values}
}}
""",
    )



# flake8: noqa: E501
import sys
from unittest import mock

import pytest
from gotranx.codegen import CCodeGenerator
from gotranx.codegen import RHSArgument
from gotranx.ode import make_ode


@pytest.fixture(scope="module")
def ode(trans, parser):
    expr = """
    parameters(a=0)
    parameters("My component",
    sigma=ScalarParam(12.0, description="Some description"),
    rho=21.0,
    beta=2.4
    )
    states("My component", x=1.0, y=2.0,z=3.05)

    expressions("My component")
    dy_dt = x*(rho - z) - y # millivolt
    dx_dt = sigma*(-x + y)
    dz_dt = -beta*z + x*y
    """
    tree = parser.parse(expr)
    return make_ode(*trans.transform(tree), name="lorentz")


@pytest.fixture(scope="module")
def codegen(ode) -> CCodeGenerator:
    return CCodeGenerator(ode)


def test_c_codegen_initial_state_values(codegen: CCodeGenerator):
    assert codegen.initial_state_values() == (
        "\nvoid init_state_values(double *states)\n"
        "{\n"
        "    /*\n"
        "    x=1.0, y=2.0, z=3.05\n"
        "    */\n"
        "    states[0] = 1.0;\n"
        "    states[1] = 2.0;\n"
        "    states[2] = 3.05;\n"
        "}\n"
    )


def test_c_codegen_initial_parameter_values(codegen: CCodeGenerator):
    assert codegen.initial_parameter_values() == (
        "\nvoid init_parameter_values(double *parameters)\n"
        "{\n"
        "    /*\n"
        "    a=0, beta=2.4, rho=21.0, sigma=12.0\n"
        "    */\n"
        "    parameters[0] = 0;\n"
        "    parameters[1] = 2.4;\n"
        "    parameters[2] = 21.0;\n"
        "    parameters[3] = 12.0;\n"
        "}\n"
    )


def test_c_codegen_initial_parameter_values_no_clang_format(ode):
    with mock.patch.dict(sys.modules, {"clang_format_docs": None}):
        codegen = CCodeGenerator(ode)
        assert codegen.initial_parameter_values() == (
            "\nvoid init_parameter_values(double* parameters){\n"
            "    /*\n"
            "    a=0, beta=2.4, rho=21.0, sigma=12.0\n"
            "    */\n"
            "    parameters[0] = 0;\n"
            "    parameters[1] = 2.4;\n"
            "    parameters[2] = 21.0;\n"
            "    parameters[3] = 12.0;\n"
            "}\n"
        )


@pytest.mark.parametrize(
    "order, arguments",
    [
        (
            RHSArgument.stp,
            (
                "const double *__restrict states, const double t, "
                "const double *__restrict parameters, double *values"
            ),
        ),
        (
            RHSArgument.spt,
            (
                "const double *__restrict states, "
                "const double *__restrict parameters, const double t, double *values"
            ),
        ),
        (
            RHSArgument.tsp,
            (
                "const double t, const double *__restrict states, "
                "const double *__restrict parameters, double *values"
            ),
        ),
    ],
)
def test_c_codegen_rhs(order: str, arguments: str, codegen: CCodeGenerator):
    assert codegen.rhs(order=order) == (
        f"\nvoid rhs({arguments})\n"
        "{\n"
        "\n"
        "    // Assign states\n"
        "    const double x = states[0];\n"
        "    const double y = states[1];\n"
        "    const double z = states[2];\n"
        "\n"
        "    // Assign parameters\n"
        "    const double a = parameters[0];\n"
        "    const double beta = parameters[1];\n"
        "    const double rho = parameters[2];\n"
        "    const double sigma = parameters[3];\n"
        "\n"
        "    // Assign expressions\n"
        "    const double dx_dt = sigma * (-x + y);\n"
        "    values[0] = dx_dt;\n"
        "    const double dy_dt = x * (rho - z) - y;\n"
        "    values[1] = dy_dt;\n"
        "    const double dz_dt = (-beta) * z + x * y;\n"
        "    values[2] = dz_dt;\n"
        "}\n"
    )


def test_c_codegen_forward_euler(codegen: CCodeGenerator):
    assert codegen.scheme("forward_euler") == (
        "\nvoid forward_euler(const double *__restrict states, const double t, const double dt,"
        "\n                   const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double x = states[0];"
        "\n    const double y = states[1];"
        "\n    const double z = states[2];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    values[0] = dt * dx_dt + x;"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    values[1] = dt * dy_dt + y;"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    values[2] = dt * dz_dt + z;"
        "\n}"
        "\n"
    )


def test_c_codegen_forward_generalized_rush_larsen(codegen: CCodeGenerator):
    assert codegen.scheme("forward_generalized_rush_larsen") == (
        "\nvoid forward_generalized_rush_larsen(const double *__restrict states, const double t, const double dt,"
        "\n                                     const double *__restrict parameters, double *values)"
        "\n{"
        "\n"
        "\n    // Assign states"
        "\n    const double x = states[0];"
        "\n    const double y = states[1];"
        "\n    const double z = states[2];"
        "\n"
        "\n    // Assign parameters"
        "\n    const double a = parameters[0];"
        "\n    const double beta = parameters[1];"
        "\n    const double rho = parameters[2];"
        "\n    const double sigma = parameters[3];"
        "\n"
        "\n    // Assign expressions"
        "\n    const double dx_dt = sigma * (-x + y);"
        "\n    const double dx_dt_linearized = -sigma;"
        "\n    values[0] = x + ((fabs(dx_dt_linearized) > 1e-08) ? (dx_dt * (exp(dt * dx_dt_linearized) - 1) / dx_dt_linearized)"
        "\n                                                      : (dt * dx_dt));"
        "\n    const double dy_dt = x * (rho - z) - y;"
        "\n    const double dy_dt_linearized = -1;"
        "\n    values[1] = y + ((fabs(dy_dt_linearized) > 1e-08) ? (dy_dt * (exp(dt * dy_dt_linearized) - 1) / dy_dt_linearized)"
        "\n                                                      : (dt * dy_dt));"
        "\n    const double dz_dt = (-beta) * z + x * y;"
        "\n    const double dz_dt_linearized = -beta;"
        "\n    values[2] = z + ((fabs(dz_dt_linearized) > 1e-08) ? (dz_dt * (exp(dt * dz_dt_linearized) - 1) / dz_dt_linearized)"
        "\n                                                      : (dt * dz_dt));"
        "\n}"
        "\n"
    )



from __future__ import annotations
from pathlib import Path

from structlog import get_logger

from . import exceptions
from .ode import make_ode, ODE
from .parser import Parser
from .transformer import TreeToODE


logger = get_logger()


def ode_from_string(text: str, name="ode") -> ODE:
    parser = Parser(parser="lalr", transformer=TreeToODE())
    result = parser.parse(text)
    ode = make_ode(
        components=result.components,
        name=name,
        comments=result.comments,
    )
    logger.info(f"Num states {ode.num_states}")
    logger.info(f"Num parameters {ode.num_parameters}")
    return ode


def load_ode(path: str | Path):
    fname = Path(path)

    logger.info(f"Load ode {path}")

    if not fname.is_file():
        raise exceptions.ODEFileNotFound(fname)

    return ode_from_string(fname.read_text(), name=fname.stem)
